import os
import torch
import time

from torchelper.models.model_builder import ModelBuilder
 
from .callback import Callback
from torchelper.models.base_model import BaseModel
from torchelper.utils.dist_util import master_only, get_bare_model
from torchelper.utils import logger

class CkptCallback(Callback):
    def __init__(self, ckpt_dir, name, restore_epoch=-1,
                 max_ckpts=10,
                 save_per_secs=2*60*60,
                 strict=False,
                 save_before_train=True,
                 best_metric_name=None):
        super().__init__()
        self.name = name
        self.best_metric_name = best_metric_name
        self.strict = strict
        self.ckpt_dir = ckpt_dir
        self.restore_epoch = restore_epoch
        self.epochs = []
        self.last_save_time = -1
        self.max_ckpts=max_ckpts
        self.save_per_secs = save_per_secs
        self.save_before_train = save_before_train
        self.best_metric = self.__find_cur_best_model()
    
    def __find_cur_best_model(self):
        names = os.listdir(self.ckpt_dir)
        cur_best = -1
        for name in names:
            if name.startswith('best_'):
                fbest = name[0:-4].split('_')[-1]
                best = float(fbest)
                if best >cur_best:
                    cur_best = best
        return cur_best

    def load_weights(self, model, epoch):
        # 1. load weights
        save_filename = "%s_weights_%s.pth" % (epoch, self.name)
        save_path = os.path.join(self.ckpt_dir, save_filename)
        if os.path.exists(save_path):
            weights = torch.load(save_path, map_location=lambda storage, loc: storage)
            weights = get_bare_model(model).remap_weights_name(weights)
            get_bare_model(model).load_state_dict(weights, strict=self.strict)
            logger.log("success load model:"+ save_path)
        else:
            logger.log("%s not exists yet!" % save_path)
            return False
        
        # 2. load optimizer
        optimizer = get_bare_model(model).get_optimizer()
        if optimizer is not None:
            save_filename = "%s_optimizer_%s.pth" % (epoch, self.name)
            save_path = os.path.join(self.ckpt_dir, save_filename)
            if not os.path.isfile(save_path):
                logger.log("%s not exists yet!" % save_path)
                return False
            else:
                weights = torch.load(save_path, map_location=lambda storage, loc: storage)
                try:
                    optimizer.load_state_dict(weights)
                except:
                    logger.log('Failed to load optimizer parameters')
                    return False
                logger.log("success load optimizer:", save_path)
        return True
    
    def __is_best(self, metric_dict:dict):
        if self.best_metric_name is None:
            return False
 
        if metric_dict is None:
            return False

        metric = metric_dict.get(self.best_metric_name, -1)
        if metric > self.best_metric:
            self.best_metric = metric
            return True
        return False

    @master_only
    def __save_if_best(self, model:BaseModel, epoch:int, metric_dict:dict):
        if not self.__is_best(metric_dict):
            return
        if self.best_metric is None:
            return
        opt_name = "best_%s_opt_%s_%.5f.pth" % (epoch, self.name, self.best_metric)
        weight_name = "best_%s_w_%s_%.5f.pth" % (epoch, self.name, self.best_metric)
        opt_path = os.path.join(self.ckpt_dir, opt_name)
        weight_path = os.path.join(self.ckpt_dir, weight_name)
        self.__save_model(model, opt_path, weight_path)
        names = os.listdir(self.ckpt_dir)
        valid_pre = "best_%s_"%(epoch)
        for name in names:
            if name.startswith('best_') and not name.startswith(valid_pre):
                os.remove(os.path.join(self.ckpt_dir, name))

    def __save_model(self, model:BaseModel, opt_path:str, weight_path:str):
        #1. save optimizer
        optimizer = model.get_optimizer()
        if optimizer is not None:
            torch.save(optimizer.state_dict(), opt_path)
            logger.log('save:', opt_path)

        #2. save weights
        torch.save(get_bare_model(model).state_dict(), weight_path)
        logger.log('save:', weight_path)

    @master_only
    def save_model(self, model:BaseModel, epoch):
        save_opt_name = "%s_optimizer_%s.pth" % (epoch, self.name)
        save_opt_path = os.path.join(self.ckpt_dir, save_opt_name)
        save_weight_filename = "%s_weights_%s.pth" % (epoch, self.name)
        save_weight_path = os.path.join(self.ckpt_dir, save_weight_filename)
        self.__save_model(model, save_opt_path, save_weight_path)

        #1. save optimizer
        # optimizer = model.get_optimizer()
        # if optimizer is not None:
        #     save_opt_name = "%s_optimizer_%s.pth" % (epoch, self.name)
        #     save_opt_path = os.path.join(self.ckpt_dir, save_opt_name)
            
        #     # torch.save(optimizer.state_dict(), save_opt_path)
        #     # print('save:', save_opt_path)

        # #2. save weights
        # save_weight_filename = "%s_weights_%s.pth" % (epoch, self.name)
        # save_weight_path = os.path.join(self.ckpt_dir, save_weight_filename)
        # torch.save(get_bare_model(model).state_dict(), save_weight_path)
        # print('save:', save_weight_path)

        #3. clear old
        if self.max_ckpts<=0: #不清除
            return
        # 每间隔max_time时间保存一次
        if time.time() - self.last_save_time > self.save_per_secs:
            self.last_save_time = time.time()
        else:
            if epoch not in self.epochs:
                self.epochs.append(epoch)
            while len(self.epochs) > self.max_ckpts:
                opt_name = '%s_optimizer_%s.pth'%(self.epochs[0], self.name)
                weight_name = '%s_weights_%s.pth'%(self.epochs[0], self.name)
                opt_path = os.path.join(self.ckpt_dir, opt_name)
                weight_path = os.path.join(self.ckpt_dir, weight_name)
                print(opt_path)
                if os.path.exists(opt_path):
                    os.remove(opt_path)
                if os.path.exists(weight_path):
                    os.remove(weight_path)
                del self.epochs[0]


    def on_begin_train(self, model:BaseModel, epoch:int):
        self.load_weights(model, self.restore_epoch)
        
        if self.save_before_train and epoch==0:
            self.save_model(model, -1)
        
    def on_end_val(self, model:BaseModel, epoch:int, metric_dict:dict):
        self.__save_if_best(model, epoch, metric_dict)

    def on_end_train(self, model:BaseModel):
        pass

    def on_begin_epoch(self, model:BaseModel, epoch:int):
        pass

    def on_end_epoch(self, model:BaseModel, epoch:int):
        self.save_model(model, epoch)

    def on_begin_step(self, model:BaseModel, epoch:int, step:int):
        pass

    def on_end_step(self, model:BaseModel, epoch:int, step:int):
        pass



import os
import cv2
import numpy as np
from skimage.metrics import structural_similarity as compare_ssim
from skimage.metrics import peak_signal_noise_ratio as compare_psnr
import torch
from torchelper.utils.dist_util import get_rank
from .measure import Metric
import lpips


class LPIPS(Metric):
    def __init__(self, gpu_id, net='alex'):
        '''
        :param net: str, 网络结构
        '''
        super().__init__()
        self.device = torch.device('cuda:'+str(gpu_id) if gpu_id is not None else 'cpu')
        self.model = lpips.LPIPS(net=net)
        self.model = self.model.to(self.device)
    
 
    def cal(self, inp, tar):
        '''测量lpips
        :param inp: np.array, dtype=np.uint8, 输入图
        :param tar: np.array, dtype=np.uint8, GT
        :return: 返回lpips值
        '''
        inp = self.img_to_batch_tensor(inp, self.device)
        tar = self.img_to_batch_tensor(tar, self.device)
        inp = inp / 127.5 - 1
        tar = tar / 127.5 - 1
        
        dist01 = self.model.forward(inp, tar).detach().cpu().numpy()
        val = np.mean(dist01)
        return val


import torch
import math
import torch.nn as nn
import itertools
from torchelper import ModelGroup, SSIMNet
from ..archs.vgg_19 import VGG19_torch
import torch.nn.functional as F
from torch.nn.parallel import gather, parallel_apply, replicate
from torch.cuda.amp import GradScaler

class OMGDDistiller(ModelGroup):

    def __init__(self, cfg, gpu_id, is_train, is_dist, is_amp):
        super().__init__(is_train, cfg['ckpt_dir'], gpu_id, is_dist, is_amp)
        self.connectors = []
        self.teacher_w = None
        self.student = None
        self.vgg = VGG19_torch().cuda(gpu_id)
    
 
    def add_student(self, teacher_w, name, cls_str, init_lr, loss_func, model_cfg, lr_scheduler):
        self.teacher_w = self.get_bare_model(teacher_w)
        self.student = self.create_net(cls_str, model_cfg)

        teacher_chans = self.teacher_w.get_mid_features()
        student_chans = self.student.get_mid_features()
        params = [self.student.parameters()]
        for t_chan, s_chan in zip(teacher_chans, student_chans):
            module = self.build_feature_connector(t_chan, s_chan)
            params.append(module.parameters())
            self.connectors.append(module.cuda(self.gpu_id))
        optimizer = torch.optim.Adam(itertools.chain(*params), lr=init_lr, betas=(0.5, 0.999))
        self.optimizers[name] = optimizer
        self.loss_funcs[name] = loss_func
        self.schedulers[name] = lr_scheduler
        if self.is_amp:
            self.amp_scaler[name] = GradScaler()
        model = self.model_to_device(self.student)
        self.models[name] = model

     

    def build_feature_connector(self, t_channel, s_channel):
        modules = [nn.Conv2d(s_channel, t_channel, kernel_size=1, stride=1, padding=0, bias=False),
             nn.BatchNorm2d(t_channel),
             nn.ReLU(inplace=True)]

        for module in modules:
            if isinstance(module, nn.Conv2d):
                n = module.kernel_size[0] * module.kernel_size[1] * module.out_channels
                module.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(module, nn.BatchNorm2d):
                module.weight.data.fill_(1)
                module.bias.data.zero_()
        return nn.Sequential(*modules)

    def calc_cd_loss(self):
        t_feats = self.teacher_w.get_mid_features()
        s_feats = self.student.get_mid_features()
        losses = []
        for connector, s_feat, target in zip(self.connectors, s_feats, t_feats):
            s_act = connector(s_feat)
            source, target = s_act, target.detach()
            source = source.mean(dim=(2, 3), keepdim=False)
            target = target.mean(dim=(2, 3), keepdim=False)
            loss = torch.mean(torch.pow(source - target, 2))
            losses.append(loss)
        return sum(losses)

    def cal_student_loss(self, fake_s, fake_t_w, fake_t_d):
        fake_ts = [fake_t_d.detach(), fake_t_w.detach()]
        lambda_ssim = 10
        lambda_style = 10
        loss_g_student = 0
        for fake_t in fake_ts:
            ssim_loss = SSIMNet()
            loss_g_ssim = (1 - ssim_loss(fake_s, fake_t)) * lambda_ssim
            t_feats = self.vgg(fake_t)
            s_feats = self.vgg(fake_s)
            t_gram = [self._gram_mat(fmap) for fmap in t_feats]
            s_gram = [self._gram_mat(fmap) for fmap in s_feats]

            loss_g_style = 0
            for i in range(len(t_gram)):
                loss_g_style += lambda_style * F.l1_loss(s_gram[i], t_gram[i])
            s_recon, t_recon = s_feats[1], t_feats[1]
            lambda_feature = 10
            loss_g_feature = lambda_feature * F.l1_loss(s_recon, t_recon)
            # diff_i = torch.sum(torch.abs(self.Sfake_B[:, :, :, 1:] - self.Sfake_B[:, :, :, :-1]))
            # diff_j = torch.sum(torch.abs(self.Sfake_B[:, :, 1:, :] - self.Sfake_B[:, :, :-1, :]))
            # self.loss_G_tv = self.lambda_tv * (diff_i + diff_j)
            loss_g_student += loss_g_ssim + loss_g_style + loss_g_feature # + self.loss_G_tv
        
        lambda_cd = 10
        loss_g_student += self.calc_cd_loss() * lambda_cd
        return loss_g_student


import numpy as np
from .measure import Metric
class PSNR(Metric):
    def __init__(self):
        super().__init__()

    def calculate_psnr(self, img1, img2):
        """Calculate PSNR (Peak Signal-to-Noise Ratio).
        Ref: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
        Args:
            img1 (ndarray): Images with range [0, 255].
            img2 (ndarray): Images with range [0, 255].
        Returns:
            float: psnr result.
        """
        img1 = img1.astype(np.float64)
        img2 = img2.astype(np.float64)
        mse = np.mean((img1 - img2)**2)
        if mse == 0:
            return float('inf')
        return 20. * np.log10(255. / np.sqrt(mse))

   
    def cal(self, inp, tar):
        if len(inp)==3:
            batch_size = 1
            inp, tar = [inp], [tar]
        else:
            batch_size = inp.shape[0]
        sum_v = 0
        for i in range(batch_size):
            sum_v += self.calculate_psnr(inp[i], tar[i])
        return sum_v * 1.0 / batch_size
    


import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models

try:
    from torchvision.models.utils import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

# Inception weights ported to Pytorch from
# http://download.tensorflow.org/models/image/imagenet/inception-2015-12-05.tgz
FID_WEIGHTS_URL = 'https://github.com/mseitzer/pytorch-fid/releases/download/fid_weights/pt_inception-2015-12-05-6726825d.pth'


class InceptionV3(nn.Module):
    """Pretrained InceptionV3 network returning feature maps"""

    # Index of default block of inception to return,
    # corresponds to output of final average pooling
    DEFAULT_BLOCK_INDEX = 3

    # Maps feature dimensionality to their output blocks indices
    BLOCK_INDEX_BY_DIM = {
        64: 0,  # First max pooling features
        192: 1,  # Second max pooling featurs
        768: 2,  # Pre-aux classifier features
        2048: 3  # Final average pooling features
    }

    def __init__(self,
                 output_blocks=[DEFAULT_BLOCK_INDEX],
                 resize_input=True,
                 normalize_input=True,
                 requires_grad=False,
                 use_fid_inception=True):
        """Build pretrained InceptionV3

        Parameters
        ----------
        output_blocks : list of int
            Indices of blocks to return features of. Possible values are:
                - 0: corresponds to output of first max pooling
                - 1: corresponds to output of second max pooling
                - 2: corresponds to output which is fed to aux classifier
                - 3: corresponds to output of final average pooling
        resize_input : bool
            If true, bilinearly resizes input to width and height 299 before
            feeding input to model. As the network without fully connected
            layers is fully convolutional, it should be able to handle inputs
            of arbitrary size, so resizing might not be strictly needed
        normalize_input : bool
            If true, scales the input from range (0, 1) to the range the
            pretrained Inception network expects, namely (-1, 1)
        requires_grad : bool
            If true, parameters of the model require gradients. Possibly useful
            for finetuning the network
        use_fid_inception : bool
            If true, uses the pretrained Inception model used in Tensorflow's
            FID implementation. If false, uses the pretrained Inception model
            available in torchvision. The FID Inception model has different
            weights and a slightly different structure from torchvision's
            Inception model. If you want to compute FID scores, you are
            strongly advised to set this parameter to true to get comparable
            results.
        """
        super(InceptionV3, self).__init__()

        self.resize_input = resize_input
        self.normalize_input = normalize_input
        self.output_blocks = sorted(output_blocks)
        self.last_needed_block = max(output_blocks)

        assert self.last_needed_block <= 3, \
            'Last possible output block index is 3'

        self.blocks = nn.ModuleList()

        if use_fid_inception:
            inception = fid_inception_v3()
        else:
            inception = models.inception_v3(pretrained=True)

        # Block 0: input to maxpool1
        block0 = [
            inception.Conv2d_1a_3x3,
            inception.Conv2d_2a_3x3,
            inception.Conv2d_2b_3x3,
            nn.MaxPool2d(kernel_size=3, stride=2)
        ]
        self.blocks.append(nn.Sequential(*block0))

        # Block 1: maxpool1 to maxpool2
        if self.last_needed_block >= 1:
            block1 = [
                inception.Conv2d_3b_1x1,
                inception.Conv2d_4a_3x3,
                nn.MaxPool2d(kernel_size=3, stride=2)
            ]
            self.blocks.append(nn.Sequential(*block1))

        # Block 2: maxpool2 to aux classifier
        if self.last_needed_block >= 2:
            block2 = [
                inception.Mixed_5b,
                inception.Mixed_5c,
                inception.Mixed_5d,
                inception.Mixed_6a,
                inception.Mixed_6b,
                inception.Mixed_6c,
                inception.Mixed_6d,
                inception.Mixed_6e,
            ]
            self.blocks.append(nn.Sequential(*block2))

        # Block 3: aux classifier to final avgpool
        if self.last_needed_block >= 3:
            block3 = [
                inception.Mixed_7a,
                inception.Mixed_7b,
                inception.Mixed_7c,
                nn.AdaptiveAvgPool2d(output_size=(1, 1))
            ]
            self.blocks.append(nn.Sequential(*block3))

        for param in self.parameters():
            param.requires_grad = requires_grad

    def forward(self, inp):
        """Get Inception feature maps

        Parameters
        ----------
        inp : torch.autograd.Variable
            Input tensor of shape Bx3xHxW. Values are expected to be in
            range (0, 1)

        Returns
        -------
        List of torch.autograd.Variable, corresponding to the selected output
        block, sorted ascending by index
        """
        outp = []
        x = inp

        if self.resize_input:
            x = F.interpolate(x,
                              size=(299, 299),
                              mode='bilinear',
                              align_corners=False)

        if self.normalize_input:
            x = 2 * x - 1  # Scale from range (0, 1) to range (-1, 1)

        for idx, block in enumerate(self.blocks):
            x = block(x)
            if idx in self.output_blocks:
                outp.append(x)

            if idx == self.last_needed_block:
                break

        return outp


def fid_inception_v3():
    """Build pretrained Inception model for FID computation

    The Inception model for FID computation uses a different set of weights
    and has a slightly different structure than torchvision's Inception.

    This method first constructs torchvision's Inception and then patches the
    necessary parts that are different in the FID Inception model.
    """
    inception = models.inception_v3(num_classes=1008,
                                    aux_logits=False,
                                    pretrained=False)
    inception.Mixed_5b = FIDInceptionA(192, pool_features=32)
    inception.Mixed_5c = FIDInceptionA(256, pool_features=64)
    inception.Mixed_5d = FIDInceptionA(288, pool_features=64)
    inception.Mixed_6b = FIDInceptionC(768, channels_7x7=128)
    inception.Mixed_6c = FIDInceptionC(768, channels_7x7=160)
    inception.Mixed_6d = FIDInceptionC(768, channels_7x7=160)
    inception.Mixed_6e = FIDInceptionC(768, channels_7x7=192)
    inception.Mixed_7b = FIDInceptionE_1(1280)
    inception.Mixed_7c = FIDInceptionE_2(2048)

    state_dict = load_state_dict_from_url(FID_WEIGHTS_URL, progress=True)
    inception.load_state_dict(state_dict)
    return inception


class FIDInceptionA(models.inception.InceptionA):
    """InceptionA block patched for FID computation"""

    def __init__(self, in_channels, pool_features):
        super(FIDInceptionA, self).__init__(in_channels, pool_features)

    def forward(self, x):
        branch1x1 = self.branch1x1(x)

        branch5x5 = self.branch5x5_1(x)
        branch5x5 = self.branch5x5_2(branch5x5)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = self.branch3x3dbl_3(branch3x3dbl)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch5x5, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionC(models.inception.InceptionC):
    """InceptionC block patched for FID computation"""

    def __init__(self, in_channels, channels_7x7):
        super(FIDInceptionC, self).__init__(in_channels, channels_7x7)

    def forward(self, x):
        branch1x1 = self.branch1x1(x)

        branch7x7 = self.branch7x7_1(x)
        branch7x7 = self.branch7x7_2(branch7x7)
        branch7x7 = self.branch7x7_3(branch7x7)

        branch7x7dbl = self.branch7x7dbl_1(x)
        branch7x7dbl = self.branch7x7dbl_2(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_3(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_4(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_5(branch7x7dbl)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch7x7, branch7x7dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionE_1(models.inception.InceptionE):
    """First InceptionE block patched for FID computation"""

    def __init__(self, in_channels):
        super(FIDInceptionE_1, self).__init__(in_channels)

    def forward(self, x):
        branch1x1 = self.branch1x1(x)

        branch3x3 = self.branch3x3_1(x)
        branch3x3 = [
            self.branch3x3_2a(branch3x3),
            self.branch3x3_2b(branch3x3),
        ]
        branch3x3 = torch.cat(branch3x3, 1)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = [
            self.branch3x3dbl_3a(branch3x3dbl),
            self.branch3x3dbl_3b(branch3x3dbl),
        ]
        branch3x3dbl = torch.cat(branch3x3dbl, 1)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionE_2(models.inception.InceptionE):
    """Second InceptionE block patched for FID computation"""

    def __init__(self, in_channels):
        super(FIDInceptionE_2, self).__init__(in_channels)

    def forward(self, x):
        branch1x1 = self.branch1x1(x)

        branch3x3 = self.branch3x3_1(x)
        branch3x3 = [
            self.branch3x3_2a(branch3x3),
            self.branch3x3_2b(branch3x3),
        ]
        branch3x3 = torch.cat(branch3x3, 1)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = [
            self.branch3x3dbl_3a(branch3x3dbl),
            self.branch3x3dbl_3b(branch3x3dbl),
        ]
        branch3x3dbl = torch.cat(branch3x3dbl, 1)

        # Patch: The FID Inception model uses max pooling instead of average
        # pooling. This is likely an error in this specific Inception
        # implementation, as other Inception models use average pooling here
        # (which matches the description in the paper).
        branch_pool = F.max_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)



import math
import random
import torch
from torch import nn
from torch.nn import functional as F

from torchelper.ops.fused_act import FusedLeakyReLU, fused_leaky_relu
from torchelper.ops.upfirdn2d import upfirdn2d


class NormStyleCode(nn.Module):

    def forward(self, x):
        """Normalize the style codes.
        Args:
            x (Tensor): Style codes with shape (b, c).
        Returns:
            Tensor: Normalized tensor.
        """
        return x * torch.rsqrt(torch.mean(x**2, dim=1, keepdim=True) + 1e-8)


def make_resample_kernel(k):
    """Make resampling kernel for UpFirDn.
    Args:
        k (list[int]): A list indicating the 1D resample kernel magnitude.
    Returns:
        Tensor: 2D resampled kernel.
    """
    k = torch.tensor(k, dtype=torch.float32)
    if k.ndim == 1:
        k = k[None, :] * k[:, None]  # to 2D kernel, outer product
    # normalize
    k /= k.sum()
    return k


class UpFirDnUpsample(nn.Module):
    """Upsample, FIR filter, and downsample (upsampole version).
    References:
    1. https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.upfirdn.html  # noqa: E501
    2. http://www.ece.northwestern.edu/local-apps/matlabhelp/toolbox/signal/upfirdn.html  # noqa: E501
    Args:
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude.
        factor (int): Upsampling scale factor. Default: 2.
    """

    def __init__(self, resample_kernel, factor=2):
        super(UpFirDnUpsample, self).__init__()
        self.kernel = make_resample_kernel(resample_kernel) * (factor**2)
        self.factor = factor

        pad = self.kernel.shape[0] - factor
        self.pad = ((pad + 1) // 2 + factor - 1, pad // 2)

    def forward(self, x):
        out = upfirdn2d(x, self.kernel.type_as(x), up=self.factor, down=1, pad=self.pad)
        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(factor={self.factor})')


class UpFirDnDownsample(nn.Module):
    """Upsample, FIR filter, and downsample (downsampole version).
    Args:
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude.
        factor (int): Downsampling scale factor. Default: 2.
    """

    def __init__(self, resample_kernel, factor=2):
        super(UpFirDnDownsample, self).__init__()
        self.kernel = make_resample_kernel(resample_kernel)
        self.factor = factor

        pad = self.kernel.shape[0] - factor
        self.pad = ((pad + 1) // 2, pad // 2)

    def forward(self, x):
        out = upfirdn2d(x, self.kernel.type_as(x), up=1, down=self.factor, pad=self.pad)
        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(factor={self.factor})')


class UpFirDnSmooth(nn.Module):
    """Upsample, FIR filter, and downsample (smooth version).
    Args:
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude.
        upsample_factor (int): Upsampling scale factor. Default: 1.
        downsample_factor (int): Downsampling scale factor. Default: 1.
        kernel_size (int): Kernel size: Deafult: 1.
    """

    def __init__(self, resample_kernel, upsample_factor=1, downsample_factor=1, kernel_size=1):
        super(UpFirDnSmooth, self).__init__()
        self.upsample_factor = upsample_factor
        self.downsample_factor = downsample_factor
        self.kernel = make_resample_kernel(resample_kernel)
        if upsample_factor > 1:
            self.kernel = self.kernel * (upsample_factor**2)

        if upsample_factor > 1:
            pad = (self.kernel.shape[0] - upsample_factor) - (kernel_size - 1)
            self.pad = ((pad + 1) // 2 + upsample_factor - 1, pad // 2 + 1)
        elif downsample_factor > 1:
            pad = (self.kernel.shape[0] - downsample_factor) + (kernel_size - 1)
            self.pad = ((pad + 1) // 2, pad // 2)
        else:
            raise NotImplementedError

    def forward(self, x):
        out = upfirdn2d(x, self.kernel.type_as(x), up=1, down=1, pad=self.pad)
        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(upsample_factor={self.upsample_factor}'
                f', downsample_factor={self.downsample_factor})')


class EqualLinear(nn.Module):
    """Equalized Linear as StyleGAN2.
    Args:
        in_channels (int): Size of each sample.
        out_channels (int): Size of each output sample.
        bias (bool): If set to ``False``, the layer will not learn an additive
            bias. Default: ``True``.
        bias_init_val (float): Bias initialized value. Default: 0.
        lr_mul (float): Learning rate multiplier. Default: 1.
        activation (None | str): The activation after ``linear`` operation.
            Supported: 'fused_lrelu', None. Default: None.
    """

    def __init__(self, in_channels, out_channels, bias=True, bias_init_val=0, lr_mul=1, activation=None):
        super(EqualLinear, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.lr_mul = lr_mul
        self.activation = activation
        if self.activation not in ['fused_lrelu', None]:
            raise ValueError(f'Wrong activation value in EqualLinear: {activation}'
                             "Supported ones are: ['fused_lrelu', None].")
        self.scale = (1 / math.sqrt(in_channels)) * lr_mul

        self.weight = nn.Parameter(torch.randn(out_channels, in_channels).div_(lr_mul))
        if bias:
            self.bias = nn.Parameter(torch.zeros(out_channels).fill_(bias_init_val))
        else:
            self.register_parameter('bias', None)

    def forward(self, x):
        if self.bias is None:
            bias = None
        else:
            bias = self.bias * self.lr_mul
        if self.activation == 'fused_lrelu':
            out = F.linear(x, self.weight * self.scale)
            out = fused_leaky_relu(out, bias)
        else:
            out = F.linear(x, self.weight * self.scale, bias=bias)
        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(in_channels={self.in_channels}, '
                f'out_channels={self.out_channels}, bias={self.bias is not None})')


class ModulatedConv2d(nn.Module):
    """Modulated Conv2d used in StyleGAN2.
    There is no bias in ModulatedConv2d.
    Args:
        in_channels (int): Channel number of the input.
        out_channels (int): Channel number of the output.
        kernel_size (int): Size of the convolving kernel.
        num_style_feat (int): Channel number of style features.
        demodulate (bool): Whether to demodulate in the conv layer.
            Default: True.
        sample_mode (str | None): Indicating 'upsample', 'downsample' or None.
            Default: None.
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude. Default: (1, 3, 3, 1).
        eps (float): A value added to the denominator for numerical stability.
            Default: 1e-8.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 num_style_feat,
                 demodulate=True,
                 sample_mode=None,
                 resample_kernel=(1, 3, 3, 1),
                 eps=1e-8):
        super(ModulatedConv2d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.demodulate = demodulate
        self.sample_mode = sample_mode
        self.eps = eps

        if self.sample_mode == 'upsample':
            self.smooth = UpFirDnSmooth(
                resample_kernel, upsample_factor=2, downsample_factor=1, kernel_size=kernel_size)
        elif self.sample_mode == 'downsample':
            self.smooth = UpFirDnSmooth(
                resample_kernel, upsample_factor=1, downsample_factor=2, kernel_size=kernel_size)
        elif self.sample_mode is None:
            pass
        else:
            raise ValueError(f'Wrong sample mode {self.sample_mode}, '
                             "supported ones are ['upsample', 'downsample', None].")

        self.scale = 1 / math.sqrt(in_channels * kernel_size**2)
        # modulation inside each modulated conv
        self.modulation = EqualLinear(
            num_style_feat, in_channels, bias=True, bias_init_val=1, lr_mul=1, activation=None)

        self.weight = nn.Parameter(torch.randn(1, out_channels, in_channels, kernel_size, kernel_size))
        self.padding = kernel_size // 2

    def forward(self, x, style):
        """Forward function.
        Args:
            x (Tensor): Tensor with shape (b, c, h, w).
            style (Tensor): Tensor with shape (b, num_style_feat).
        Returns:
            Tensor: Modulated tensor after convolution.
        """
        b, c, h, w = x.shape  # c = c_in
        # weight modulation
        style = self.modulation(style).view(b, 1, c, 1, 1)
        # self.weight: (1, c_out, c_in, k, k); style: (b, 1, c, 1, 1)
        weight = self.scale * self.weight * style  # (b, c_out, c_in, k, k)

        if self.demodulate:
            demod = torch.rsqrt(weight.pow(2).sum([2, 3, 4]) + self.eps)
            weight = weight * demod.view(b, self.out_channels, 1, 1, 1)

        weight = weight.view(b * self.out_channels, c, self.kernel_size, self.kernel_size)

        if self.sample_mode == 'upsample':
            x = x.view(1, b * c, h, w)
            weight = weight.view(b, self.out_channels, c, self.kernel_size, self.kernel_size)
            weight = weight.transpose(1, 2).reshape(b * c, self.out_channels, self.kernel_size, self.kernel_size)
            out = F.conv_transpose2d(x, weight, padding=0, stride=2, groups=b)
            out = out.view(b, self.out_channels, *out.shape[2:4])
            out = self.smooth(out)
        elif self.sample_mode == 'downsample':
            x = self.smooth(x)
            x = x.view(1, b * c, *x.shape[2:4])
            out = F.conv2d(x, weight, padding=0, stride=2, groups=b)
            out = out.view(b, self.out_channels, *out.shape[2:4])
        else:
            x = x.view(1, b * c, h, w)
            # weight: (b*c_out, c_in, k, k), groups=b
            out = F.conv2d(x, weight, padding=self.padding, groups=b)
            out = out.view(b, self.out_channels, *out.shape[2:4])

        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(in_channels={self.in_channels}, '
                f'out_channels={self.out_channels}, '
                f'kernel_size={self.kernel_size}, '
                f'demodulate={self.demodulate}, sample_mode={self.sample_mode})')


class StyleConv(nn.Module):
    """Style conv.
    Args:
        in_channels (int): Channel number of the input.
        out_channels (int): Channel number of the output.
        kernel_size (int): Size of the convolving kernel.
        num_style_feat (int): Channel number of style features.
        demodulate (bool): Whether demodulate in the conv layer. Default: True.
        sample_mode (str | None): Indicating 'upsample', 'downsample' or None.
            Default: None.
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude. Default: (1, 3, 3, 1).
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 num_style_feat,
                 demodulate=True,
                 sample_mode=None,
                 resample_kernel=(1, 3, 3, 1)):
        super(StyleConv, self).__init__()
        self.modulated_conv = ModulatedConv2d(
            in_channels,
            out_channels,
            kernel_size,
            num_style_feat,
            demodulate=demodulate,
            sample_mode=sample_mode,
            resample_kernel=resample_kernel)
        self.weight = nn.Parameter(torch.zeros(1))  # for noise injection
        self.activate = FusedLeakyReLU(out_channels)

    def forward(self, x, style, noise=None):
        # modulate
        out = self.modulated_conv(x, style)
        # noise injection
        if noise is None:
            b, _, h, w = out.shape
            noise = out.new_empty(b, 1, h, w).normal_()
        out = out + self.weight * noise
        # activation (with bias)
        out = self.activate(out)
        return out


class ToRGB(nn.Module):
    """To RGB from features.
    Args:
        in_channels (int): Channel number of input.
        num_style_feat (int): Channel number of style features.
        upsample (bool): Whether to upsample. Default: True.
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude. Default: (1, 3, 3, 1).
    """

    def __init__(self, in_channels, num_style_feat, upsample=True, resample_kernel=(1, 3, 3, 1)):
        super(ToRGB, self).__init__()
        if upsample:
            self.upsample = UpFirDnUpsample(resample_kernel, factor=2)
        else:
            self.upsample = None
        self.modulated_conv = ModulatedConv2d(
            in_channels, 3, kernel_size=1, num_style_feat=num_style_feat, demodulate=False, sample_mode=None)
        self.bias = nn.Parameter(torch.zeros(1, 3, 1, 1))

    def forward(self, x, style, skip=None):
        """Forward function.
        Args:
            x (Tensor): Feature tensor with shape (b, c, h, w).
            style (Tensor): Tensor with shape (b, num_style_feat).
            skip (Tensor): Base/skip tensor. Default: None.
        Returns:
            Tensor: RGB images.
        """
        out = self.modulated_conv(x, style)
        out = out + self.bias
        if skip is not None:
            if self.upsample:
                skip = self.upsample(skip)
            out = out + skip
        return out


class ConstantInput(nn.Module):
    """Constant input.
    Args:
        num_channel (int): Channel number of constant input.
        size (int): Spatial size of constant input.
    """

    def __init__(self, num_channel, size):
        super(ConstantInput, self).__init__()
        self.weight = nn.Parameter(torch.randn(1, num_channel, size, size))

    def forward(self, batch):
        out = self.weight.repeat(batch, 1, 1, 1)
        return out

 
class StyleGAN2Generator(nn.Module):
    """StyleGAN2 Generator.
    Args:
        out_size (int): The spatial size of outputs.
        num_style_feat (int): Channel number of style features. Default: 512.
        num_mlp (int): Layer number of MLP style layers. Default: 8.
        channel_multiplier (int): Channel multiplier for large networks of
            StyleGAN2. Default: 2.
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude. A cross production will be applied to extent 1D resample
            kenrel to 2D resample kernel. Default: (1, 3, 3, 1).
        lr_mlp (float): Learning rate multiplier for mlp layers. Default: 0.01.
        narrow (float): Narrow ratio for channels. Default: 1.0.
    """

    def __init__(self,
                 out_size,
                 num_style_feat=512,
                 num_mlp=8,
                 channel_multiplier=2,
                 resample_kernel=(1, 3, 3, 1),
                 lr_mlp=0.01,
                 narrow=1):
        super(StyleGAN2Generator, self).__init__()
        # Style MLP layers
        self.num_style_feat = num_style_feat
        style_mlp_layers = [NormStyleCode()]
        for i in range(num_mlp):
            style_mlp_layers.append(
                EqualLinear(
                    num_style_feat, num_style_feat, bias=True, bias_init_val=0, lr_mul=lr_mlp,
                    activation='fused_lrelu'))
        self.style_mlp = nn.Sequential(*style_mlp_layers)

        channels = {
            '4': int(512 * narrow),
            '8': int(512 * narrow),
            '16': int(512 * narrow),
            '32': int(512 * narrow),
            '64': int(256 * channel_multiplier * narrow),
            '128': int(128 * channel_multiplier * narrow),
            '256': int(64 * channel_multiplier * narrow),
            '512': int(32 * channel_multiplier * narrow),
            '1024': int(16 * channel_multiplier * narrow)
        }
        self.channels = channels

        self.constant_input = ConstantInput(channels['4'], size=4)
        self.style_conv1 = StyleConv(
            channels['4'],
            channels['4'],
            kernel_size=3,
            num_style_feat=num_style_feat,
            demodulate=True,
            sample_mode=None,
            resample_kernel=resample_kernel)
        self.to_rgb1 = ToRGB(channels['4'], num_style_feat, upsample=False, resample_kernel=resample_kernel)

        self.log_size = int(math.log(out_size, 2))
        self.num_layers = (self.log_size - 2) * 2 + 1
        self.num_latent = self.log_size * 2 - 2

        self.style_convs = nn.ModuleList()
        self.to_rgbs = nn.ModuleList()
        self.noises = nn.Module()

        in_channels = channels['4']
        # noise
        for layer_idx in range(self.num_layers):
            resolution = 2**((layer_idx + 5) // 2)
            shape = [1, 1, resolution, resolution]
            self.noises.register_buffer(f'noise{layer_idx}', torch.randn(*shape))
        # style convs and to_rgbs
        for i in range(3, self.log_size + 1):
            out_channels = channels[f'{2**i}']
            self.style_convs.append(
                StyleConv(
                    in_channels,
                    out_channels,
                    kernel_size=3,
                    num_style_feat=num_style_feat,
                    demodulate=True,
                    sample_mode='upsample',
                    resample_kernel=resample_kernel,
                ))
            self.style_convs.append(
                StyleConv(
                    out_channels,
                    out_channels,
                    kernel_size=3,
                    num_style_feat=num_style_feat,
                    demodulate=True,
                    sample_mode=None,
                    resample_kernel=resample_kernel))
            self.to_rgbs.append(ToRGB(out_channels, num_style_feat, upsample=True, resample_kernel=resample_kernel))
            in_channels = out_channels

    def make_noise(self):
        """Make noise for noise injection."""
        device = self.constant_input.weight.device
        noises = [torch.randn(1, 1, 4, 4, device=device)]

        for i in range(3, self.log_size + 1):
            for _ in range(2):
                noises.append(torch.randn(1, 1, 2**i, 2**i, device=device))

        return noises

    def get_latent(self, x):
        return self.style_mlp(x)

    def mean_latent(self, num_latent):
        latent_in = torch.randn(num_latent, self.num_style_feat, device=self.constant_input.weight.device)
        latent = self.style_mlp(latent_in).mean(0, keepdim=True)
        return latent

    def forward(self,
                styles,
                input_is_latent=False,
                noise=None,
                randomize_noise=True,
                truncation=1,
                truncation_latent=None,
                inject_index=None,
                return_latents=False):
        """Forward function for StyleGAN2Generator.
        Args:
            styles (list[Tensor]): Sample codes of styles.
            input_is_latent (bool): Whether input is latent style.
                Default: False.
            noise (Tensor | None): Input noise or None. Default: None.
            randomize_noise (bool): Randomize noise, used when 'noise' is
                False. Default: True.
            truncation (float): TODO. Default: 1.
            truncation_latent (Tensor | None): TODO. Default: None.
            inject_index (int | None): The injection index for mixing noise.
                Default: None.
            return_latents (bool): Whether to return style latents.
                Default: False.
        """
        # style codes -> latents with Style MLP layer
        if not input_is_latent:
            styles = [self.style_mlp(s) for s in styles]
        # noises
        if noise is None:
            if randomize_noise:
                noise = [None] * self.num_layers  # for each style conv layer
            else:  # use the stored noise
                noise = [getattr(self.noises, f'noise{i}') for i in range(self.num_layers)]
        # style truncation
        if truncation < 1:
            style_truncation = []
            for style in styles:
                style_truncation.append(truncation_latent + truncation * (style - truncation_latent))
            styles = style_truncation
        # get style latent with injection
        if len(styles) == 1:
            inject_index = self.num_latent

            if styles[0].ndim < 3:
                # repeat latent code for all the layers
                latent = styles[0].unsqueeze(1).repeat(1, inject_index, 1)
            else:  # used for encoder with different latent code for each layer
                latent = styles[0]
        elif len(styles) == 2:  # mixing noises
            if inject_index is None:
                inject_index = random.randint(1, self.num_latent - 1)
            latent1 = styles[0].unsqueeze(1).repeat(1, inject_index, 1)
            latent2 = styles[1].unsqueeze(1).repeat(1, self.num_latent - inject_index, 1)
            latent = torch.cat([latent1, latent2], 1)

        # main generation
        out = self.constant_input(latent.shape[0])
        out = self.style_conv1(out, latent[:, 0], noise=noise[0])
        skip = self.to_rgb1(out, latent[:, 1])

        i = 1
        for conv1, conv2, noise1, noise2, to_rgb in zip(self.style_convs[::2], self.style_convs[1::2], noise[1::2],
                                                        noise[2::2], self.to_rgbs):
            out = conv1(out, latent[:, i], noise=noise1)
            out = conv2(out, latent[:, i + 1], noise=noise2)
            skip = to_rgb(out, latent[:, i + 2], skip)
            i += 2

        image = skip

        if return_latents:
            return image, latent
        else:
            return image, None


class ScaledLeakyReLU(nn.Module):
    """Scaled LeakyReLU.
    Args:
        negative_slope (float): Negative slope. Default: 0.2.
    """

    def __init__(self, negative_slope=0.2):
        super(ScaledLeakyReLU, self).__init__()
        self.negative_slope = negative_slope

    def forward(self, x):
        out = F.leaky_relu(x, negative_slope=self.negative_slope)
        return out * math.sqrt(2)


class EqualConv2d(nn.Module):
    """Equalized Linear as StyleGAN2.
    Args:
        in_channels (int): Channel number of the input.
        out_channels (int): Channel number of the output.
        kernel_size (int): Size of the convolving kernel.
        stride (int): Stride of the convolution. Default: 1
        padding (int): Zero-padding added to both sides of the input.
            Default: 0.
        bias (bool): If ``True``, adds a learnable bias to the output.
            Default: ``True``.
        bias_init_val (float): Bias initialized value. Default: 0.
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, bias=True, bias_init_val=0):
        super(EqualConv2d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.scale = 1 / math.sqrt(in_channels * kernel_size**2)

        self.weight = nn.Parameter(torch.randn(out_channels, in_channels, kernel_size, kernel_size))
        if bias:
            self.bias = nn.Parameter(torch.zeros(out_channels).fill_(bias_init_val))
        else:
            self.register_parameter('bias', None)

    def forward(self, x):
        out = F.conv2d(
            x,
            self.weight * self.scale,
            bias=self.bias,
            stride=self.stride,
            padding=self.padding,
        )

        return out

    def __repr__(self):
        return (f'{self.__class__.__name__}(in_channels={self.in_channels}, '
                f'out_channels={self.out_channels}, '
                f'kernel_size={self.kernel_size},'
                f' stride={self.stride}, padding={self.padding}, '
                f'bias={self.bias is not None})')


class ConvLayer(nn.Sequential):
    """Conv Layer used in StyleGAN2 Discriminator.
    Args:
        in_channels (int): Channel number of the input.
        out_channels (int): Channel number of the output.
        kernel_size (int): Kernel size.
        downsample (bool): Whether downsample by a factor of 2.
            Default: False.
        resample_kernel (list[int]): A list indicating the 1D resample
            kernel magnitude. A cross production will be applied to
            extent 1D resample kenrel to 2D resample kernel.
            Default: (1, 3, 3, 1).
        bias (bool): Whether with bias. Default: True.
        activate (bool): Whether use activateion. Default: True.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 downsample=False,
                 resample_kernel=(1, 3, 3, 1),
                 bias=True,
                 activate=True):
        layers = []
        # downsample
        if downsample:
            layers.append(
                UpFirDnSmooth(resample_kernel, upsample_factor=1, downsample_factor=2, kernel_size=kernel_size))
            stride = 2
            self.padding = 0
        else:
            stride = 1
            self.padding = kernel_size // 2
        # conv
        layers.append(
            EqualConv2d(
                in_channels, out_channels, kernel_size, stride=stride, padding=self.padding, bias=bias
                and not activate))
        # activation
        if activate:
            if bias:
                layers.append(FusedLeakyReLU(out_channels))
            else:
                layers.append(ScaledLeakyReLU(0.2))

        super(ConvLayer, self).__init__(*layers)


class ResBlock(nn.Module):
    """Residual block used in StyleGAN2 Discriminator.
    Args:
        in_channels (int): Channel number of the input.
        out_channels (int): Channel number of the output.
        resample_kernel (list[int]): A list indicating the 1D resample
            kernel magnitude. A cross production will be applied to
            extent 1D resample kenrel to 2D resample kernel.
            Default: (1, 3, 3, 1).
    """

    def __init__(self, in_channels, out_channels, resample_kernel=(1, 3, 3, 1)):
        super(ResBlock, self).__init__()

        self.conv1 = ConvLayer(in_channels, in_channels, 3, bias=True, activate=True)
        self.conv2 = ConvLayer(
            in_channels, out_channels, 3, downsample=True, resample_kernel=resample_kernel, bias=True, activate=True)
        self.skip = ConvLayer(
            in_channels, out_channels, 1, downsample=True, resample_kernel=resample_kernel, bias=False, activate=False)

    def forward(self, x):
        out = self.conv1(x)
        out = self.conv2(out)
        skip = self.skip(x)
        out = (out + skip) / math.sqrt(2)
        return out

 
class StyleGAN2Discriminator(nn.Module):
    """StyleGAN2 Discriminator.
    Args:
        out_size (int): The spatial size of outputs.
        channel_multiplier (int): Channel multiplier for large networks of
            StyleGAN2. Default: 2.
        resample_kernel (list[int]): A list indicating the 1D resample kernel
            magnitude. A cross production will be applied to extent 1D resample
            kenrel to 2D resample kernel. Default: (1, 3, 3, 1).
        stddev_group (int): For group stddev statistics. Default: 4.
        narrow (float): Narrow ratio for channels. Default: 1.0.
    """

    def __init__(self, out_size, channel_multiplier=2, resample_kernel=(1, 3, 3, 1), stddev_group=4, narrow=1):
        super(StyleGAN2Discriminator, self).__init__()

        channels = {
            '4': int(512 * narrow),
            '8': int(512 * narrow),
            '16': int(512 * narrow),
            '32': int(512 * narrow),
            '64': int(256 * channel_multiplier * narrow),
            '128': int(128 * channel_multiplier * narrow),
            '256': int(64 * channel_multiplier * narrow),
            '512': int(32 * channel_multiplier * narrow),
            '1024': int(16 * channel_multiplier * narrow)
        }

        log_size = int(math.log(out_size, 2))

        conv_body = [ConvLayer(3, channels[f'{out_size}'], 1, bias=True, activate=True)]

        in_channels = channels[f'{out_size}']
        for i in range(log_size, 2, -1):
            out_channels = channels[f'{2**(i - 1)}']
            conv_body.append(ResBlock(in_channels, out_channels, resample_kernel))
            in_channels = out_channels
        self.conv_body = nn.Sequential(*conv_body)

        self.final_conv = ConvLayer(in_channels + 1, channels['4'], 3, bias=True, activate=True)
        self.final_linear = nn.Sequential(
            EqualLinear(
                channels['4'] * 4 * 4, channels['4'], bias=True, bias_init_val=0, lr_mul=1, activation='fused_lrelu'),
            EqualLinear(channels['4'], 1, bias=True, bias_init_val=0, lr_mul=1, activation=None),
        )
        self.stddev_group = stddev_group
        self.stddev_feat = 1

    def forward(self, x):
        out = self.conv_body(x)

        b, c, h, w = out.shape
        # concatenate a group stddev statistics to out
        group = min(b, self.stddev_group)  # Minibatch must be divisible by (or smaller than) group_size
        stddev = out.view(group, -1, self.stddev_feat, c // self.stddev_feat, h, w)
        stddev = torch.sqrt(stddev.var(0, unbiased=False) + 1e-8)
        stddev = stddev.mean([2, 3, 4], keepdims=True).squeeze(2)
        stddev = stddev.repeat(group, 1, h, w)
        out = torch.cat([out, stddev], 1)

        out = self.final_conv(out)
        out = out.view(b, -1)
        out = self.final_linear(out)

        return out


import os
import torch
import time
import numpy as np
from .callback import Callback
from torchelper.models.base_model import BaseModel
from torchelper.utils.dist_util import master_only, get_bare_model

class ReduceLROnPlateau(Callback):
    """当指定指标没有提升时，自动下降学习率。
    patience个epoch后，如果指标没提升，将学习率调整为`CurLR = CurLR * factor`.

    示例:
    ```python
    class ExampleModel(BaseModel):
        ...
        def get_metric_dict(self):
            return {'ex_loss', self.ex_loss}
        ...

    model = ExampleModel(...)
    model.add_cb(ReduceLROnPlateau(monitor='ex_loss', factor=0.2,
                                    patience=5, min_lr=0.001))
    ```
    Args:
        monitor: str, 监控的指标名称，必须是对应Model的get_metric函数返回字典中包括的key
        factor: float, 学习率下降倍数，`new_lr = lr * factor`.
        patience: int, 指定指标没有提升的epoch次数，超过此次数学习率将会下降。
        mode: str, 模式，只能取`{'min', 'max'}`中一个. 在`'min'` 模式下, monitor的
            值越低代表指标性能越好，在`'max'`模式下，monitor值越大代表性能指标越好。
        min_delta: int, 表示阈值， 如果两个epoch对应的指标值处于这个阈值范围之内，表示指标没有提升。
        cooldown: int, cooldown个epoch为冷却期， 冷却期间不进行性能指标对比和学习率下降。
        min_lr: lr的最小衰减。 如果新旧lr之间的差异小于eps，则忽略更新。 默认值：1e-8。
    """
    def __init__(self,
               monitor,
               factor=0.1,
               patience=10,
               mode='min',
               min_delta=1e-4,
               cooldown=0,
               min_lr=0):
        
        super().__init__()
        self.monitor = monitor
        self.factor = factor
        self.min_lr = min_lr
        self.min_delta = min_delta
        self.patience = patience
        self.cooldown = cooldown
        self.cooldown_counter = 0  # Cooldown counter.
        self.wait = 0
        self.best = 0
        self.mode = mode
        self.monitor_op = None
        self._reset()
    
    def _reset(self):
        """重置等待次数和冷却次数
        """
        if self.mode not in ['min', 'max']:
            print('Learning rate reduction mode %s is unknown')
            exit()
        if (self.mode == 'min'):
            self.monitor_op = lambda a, b: np.less(a, b - self.min_delta)
            self.best = np.Inf
        else:
            self.monitor_op = lambda a, b: np.greater(a, b + self.min_delta)
            self.best = -np.Inf
        self.cooldown_counter = 0
        self.wait = 0

    def get_lr(self, model:BaseModel):
        optimizer = model.get_optimizer()
        if optimizer is not None:
            for param_group in optimizer.param_groups:
                return param_group['lr']

    def set_lr(self, model:BaseModel, lr:float):
        optimizer = model.get_optimizer()
        if optimizer is not None:
            for param_group in optimizer.param_groups:
                param_group['lr'] = lr

    def on_begin_train(self, model:BaseModel, epoch:int):
        self._reset()
    
    def on_end_epoch(self, model:BaseModel, epoch:int):
        metric:dict = model.get_metric_dict()
        current = metric.get(self.monitor, None)
        if current is None:
            print('Learning rate reduction is conditioned on metric `%s` '
                        'which is not available. Available metrics are: %s',
                        self.monitor, ','.join(list(metric.keys())))

        else:
            if self.in_cooldown():
                self.cooldown_counter -= 1
                self.wait = 0

            if self.monitor_op(current, self.best):
                self.best = current
                self.wait = 0
            elif not self.in_cooldown():
                self.wait += 1
                if self.wait >= self.patience:
                    old_lr = self.get_lr(model)
                    if old_lr is None:
                        print('can not get current learn rate !!!')
                        exit()
                    if old_lr > np.float32(self.min_lr):
                        new_lr = old_lr * self.factor
                        new_lr = max(new_lr, self.min_lr)
                        self.set_lr(model, new_lr)
                        print('\nEpoch %05d: ReduceLROnPlateau reducing learning '
                                    'rate to %s.' % (epoch + 1, new_lr))
                        self.cooldown_counter = self.cooldown
                        self.wait = 0

    def in_cooldown(self):
        return self.cooldown_counter > 0




from torchelper.models.base_model import BaseModel

class Callback():
    def __init__(self):
        pass

    def on_begin_train(self, model:BaseModel, epoch:int):
        pass

    def on_end_train(self, model:BaseModel):
        pass

    def on_begin_epoch(self, model:BaseModel, epoch:int):
        pass

    def on_end_epoch(self, model:BaseModel, epoch:int):
        pass

    def on_begin_step(self, model:BaseModel, epoch:int, step:int):
        pass

    def on_end_step(self, model:BaseModel, epoch:int, step:int):
        pass

    def on_end_val(self, model:BaseModel, epoch:int, metric_dict:dict):
        pass



import os
import cv2
from abc import ABCMeta, abstractmethod
from .base_dataset import BaseDataset

class DirDataset(BaseDataset):
    '''单目录数据读取
    '''
    def __init__(self, dir_path, dst_wh=None, img_dir=None, label_name=None, cfg=None):
        self.data_root = dir_path
        self.cfg = cfg
        self.img_dir = img_dir
        self.label_name = label_name
        self.dst_wh = dst_wh
        self.labels = {}
        self.load_data()
        self.size = len(self.names)

    def load_data(self):
        #加载label
        if self.label_name is not None:
            self.labels = self.parse_label(os.path.join(self.data_root , self.label_name))
        if self.img_dir is not None:
            self.data_root = os.path.join(self.data_root, self.img_dir)
        self.names = sorted(os.listdir(self.data_root))

    def process_img(self, name, img, label):
        return name, img, label

    def __getitem__(self, index):
        name = self.names[index]
        label = self.labels.get(name, None)
        img_path = os.path.join(self.data_root,  name)
        img = self.imread(img_path)
        if self.dst_wh is not None:
            if isinstance(self.dst_wh, list):
                img = cv2.resize(img, self.dst_wh)
            else:
                img = cv2.resize(img, (self.dst_wh, self.dst_wh))
        return self.process_img(name, img, label)

    def __len__(self):
        return self.size


class MultiDirDataset(BaseDataset, metaclass=ABCMeta):
    '''多目录读取
    '''

    def __init__(self, root_dir_path, sub_dirs=None, img_dir=None, label_name=None, dst_wh=None, cfg=None):
        self.data_root = root_dir_path
        self.img_dir = img_dir
        self.sub_dirs = sub_dirs
        self.label_name = label_name
        self.cfg = cfg
        self.dst_wh = dst_wh
        self.labels = {}
        self.names = []
        self.load_data()
        self.size = len(self.names)

    def load_data(self):
        if self.sub_dirs is None:
            dirs = os.listdir(self.data_root)
        else:
            dirs = self.sub_dirs
        for dir in dirs:
            dir_path = os.path.join(self.data_root, dir)
            if not os.path.isdir(dir_path):
                continue
            pre_name, img_dir = dir, dir_path
            #判断目录下是否有图片子目录
            if self.img_dir is not None:
                img_dir = os.path.join(img_dir, self.img_dir)
                pre_name = pre_name + '/' + self.img_dir
            #获取图片子目录+名称
            for name in os.listdir(img_dir):
                if self.is_img_name(name):
                    self.names.append(pre_name + '/' + name)
            #加载label
            if self.label_name is not None:
                self.labels = self.parse_label(self.labels, dir, os.path.join(dir_path, self.label_name))


    @abstractmethod
    def parse_label(self, label_dict, dir_name, label_path):
        '''解析label
        :param label_dict: 待更新的label字典，注意：key为目录+图片名称，例如：dir_name/imgs/img1.png，
                因为根目录是通用的，为了节约空间，省去根目录。
        :param dir_name: 目录名称
        :param label_path: label路径
        :return: dict, 返回更新后的字典
        '''
        return label_dict

    def process_img(self, name, img, label):
        return name, img, label

    def __getitem__(self, index):

        name = self.names[index]
        label = self.labels.get(name, None)
        img_path = os.path.join(self.data_root,  name)
        # img is RGB
        img = self.imread(img_path)
        if self.dst_wh is not None:
            if isinstance(self.dst_wh, list):
                img = cv2.resize(img, self.dst_wh)
            else:
                img = cv2.resize(img, (self.dst_wh, self.dst_wh))
        return self.process_img(name, img, label)

    def __len__(self):
        return self.size




def get_cls(class_str):
    '''动态创建对象
    :param class_str: str, 类名称
    '''
    arr = class_str.split('.')
    g_module = __import__('.'.join(arr[0:-1]) , fromlist = True)
    g_cls = getattr(g_module, arr[-1])
    return g_cls

def new_cls(class_str, *arg):
    '''动态创建对象
    :param class_str: str, 类名称
    :return: Model, class_str对应的模型类
    '''
    g_cls = get_cls(class_str)
    return g_cls(*arg)

def auto_new_cls(cfg:dict):
    """自动创建类，从cfg中读取cls类，并将cfg作为参数
    """
    return new_cls(cfg['cls'], *[cfg])


import os
import torch
import cv2
import numpy as np
from torch.utils.data import Dataset
from torch.utils.data.distributed import DistributedSampler
def get_data_loader(batch_size:int, dataset:Dataset, num_workers:int=0, pin_memory=True, dist:bool=True):
    '''封装DataLoader
    :param batch_size: int
    :param dataset: Dataset
    :param num_workers: int
    :param dist: bool, 是否是分布式采集数据
    '''
    sampler = None
    if dist:
        sampler = DistributedSampler(dataset) # 这个sampler会自动分配数据到各个gpu上
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
        sampler=sampler,
        drop_last=True,   # 多余的部分去除
    )
    return dataloader

class BaseDataset(Dataset):
    def imread(self, path, is_rgb:bool=True):
        '''读取图片数据
        :param is_rgb: bool, 返回格式是否是rgb
        :return: np.ndarray, dtype=np.uint8,返回图片数据
        '''
        img = cv2.imread(path)
        if img is None:
            return None
        #取前3个通道
        img = img[:, :, 0:3]
        if is_rgb:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img

    def imwrite(self, img, path, is_rgb:bool=True):
        if is_rgb:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        cv2.imwrite(path, img)

    def norm(self, uint8_img:np.ndarray, low:float=-1.0, high:float=1.0):
        '''归一化
        :param uint8_img: dtype=np.uint8, 输入图像数据
        :param low: 归一化后的最小值
        :param high: 归一化后的最大值
        :return: 归一化后的数据，数据类型为float32
        '''
        assert high > low, "high must be greater than low"
        data = uint8_img.astype(np.float32)
        data = data / 255.0 * (high - low) + low
        return data.astype(np.float32)

    def unnorm(self, float_img:np.ndarray, low:float=-1, high:float=1.0):
        '''逆归一化
        :param uint8_img: dtype=np.uint8, 输入图像数据
        :param low: 归一化后的最小值
        :param high: 归一化后的最大值
        :return: 逆归一化后的数据，数据类型为uint8
        '''
        assert high > low, "high must be greater than low"
        data = float_img /  (high - low) * 255
        data = np.clip(data, 0, 255).astype(np.uint8)
        return data

    def to_tensor(self, im_hwc):
        '''将numpy.ndarray的hwc排列的图片转为tensor
        :param im_hwc: numpy.ndarray的hwc排列的图片
        :return: tenorsor
        '''
        im_hwc = im_hwc.astype(np.float32)
        img = im_hwc.transpose((2, 0, 1))
        img = torch.from_numpy(img)
        return img

    def is_img_name(self, name):
        '''判断名称是否是图片类型
        :param name: str, 名称
        :return: bool, 判断结果
        '''
        lower = name.lower()
        return lower.endswith('.png') or \
               lower.endswith('.jpg') or \
               lower.endswith('.jpeg')



import os
import time
import copy
# from abc import ABCMeta, abstractmethod
import torch
import torch.nn as nn
from torchelper.utils.dist_util import get_bare_model, master_only
from torchelper.models import lr_scheduler as lr_scheduler
from torch.nn.parallel import DataParallel, DistributedDataParallel
from torch.cuda.amp import autocast, GradScaler
from torchelper.utils.dist_util import get_rank
from torchelper.models.base_model import BaseModel

class ModelBuilder():
    def __init__(self, is_train, ckpt_dir):
        super().__init__()
        self.dataset = None
        self.ckpt_dir = ckpt_dir
        self.is_train = is_train
        self.gpu_id = get_rank()
        if self.gpu_id>=0:
            self.device = torch.device('cuda:'+str(self.gpu_id))
        else:
            self.device = torch.device('cpu')
        self.models = {}
        self.amp_scaler = {}
        # 默认只在GPU训练
        torch.cuda.set_device(self.gpu_id)
   
    def perform_cb(self, event:str, **args):
        for __, model in self.models.items():
            self.get_bare_model(model).perform_cb(event, **args)

    # def get_print_metric(self):
    #     metric_dicts = {}
    #     for __,model in self.models.items():
    #         model = get_bare_model(model)
    #         metric_dic = model.get_metric_dict()
    #         if metric_dic is not None and isinstance(metric_dic, dict):
    #             metric_dicts.update(metric_dic)
    #     msg = []
    #     for k,v in metric_dicts.items():
    #         if v is not None:
    #             msg.append('%s:%.5f' % (k, v))

    #     return ', '.join(msg)
    
    def forward_model(self, model_name, data):
        if model_name in self.amp_scaler.keys():
            with autocast():
                out = self.models[model_name](data)
        else:
            out = self.models[model_name](data)
        return out

    def forward_wrapper(self, epoch, step, data):
        '''一个step执行前向
        :param epoch: int, 当前epoch
        :param step: int, 在当前epoch中的step数
        :param data: 训练集dataset返回的item
        '''
        self.forward(epoch, step, data)
        # if self.is_amp:
        #     # 前向过程开启 autocast
        #         self.forward(epoch, step, data)
        # else:
        #     self.forward(epoch, step, data)

    def on_begin_forward(self, data, epoch, step):
        pass
    def on_end_forward(self,  epoch, step):
        pass
    def on_begin_backward(self,  epoch, step):
        pass
    def on_end_backward(self, epoch, step):
        pass
    
    def backward_wrapper(self):
        #半浮点和全浮点的在各自函数内自定判断
        self.amp_backward()
        self.backward()
 
    def forward(self, epoch, step, data):
        pass
     
    def amp_backward(self):
        '''混合精度训练一个step执行反向
        '''
        for key, model in self.models.items():
            # loss = self.loss_dict.get(key, None) 
            if not key in self.amp_scaler.keys(): #全浮点的走全浮点的反向传播
                continue
            with autocast():
                # loss = self.get_bare_model(model).get_loss()
                loss = self.get_loss(key, self.get_bare_model(model))
            if loss is not None:
                optimizer = self.get_bare_model(model).get_optimizer()
                if optimizer is None:
                    continue
                scaler = self.amp_scaler[key]
                optimizer.zero_grad()
                # Scales loss. 为了梯度放大.
                scaler.scale(loss).backward()
                # scaler.step() 首先把梯度的值unscale回来.
                # 如果梯度的值不是 infs 或者 NaNs, 那么调用optimizer.step()来更新权重,
                # 否则，忽略step调用，从而保证权重不更新（不被破坏）
                scaler.step(optimizer)
                # 准备着，看是否要增大scaler
                scaler.update()
        
        # self.step_ema(0.5**(32 / (10 * 1000)))

    def get_loss(self, name, model):
        loss = model.get_loss()
        return loss

    def backward(self):
        '''一个step执行反向
        '''
        for key, model in self.models.items():
            if key in self.amp_scaler.keys(): #半浮点的走半浮点的反向传播
                continue
            # loss = self.loss_dict.get(key, None)
            loss = self.get_loss(key, self.get_bare_model(model))
            loss = loss.float()
            if loss is not None:
                optimizer = self.get_bare_model(model).get_optimizer()
                if optimizer is None:
                    continue
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

          
    def add_model(self, name:str, model:BaseModel):
        model.builder = self
        self.models[name] = self.model_to_device(model)
        is_half = model.cfg.get('amp', False)
        if is_half:
            self.amp_scaler[name] = GradScaler()
        

    def model_to_device(self, net:nn.Module, find_unused_parameters=False):
        """Model to device. It also warps models with DistributedDataParallel.
        :param net: nn.Module
        """
        net = net.cuda(self.gpu_id)
        net = DistributedDataParallel(
            net, device_ids=[self.gpu_id], find_unused_parameters=find_unused_parameters)
        return net

    def set_dataset(self, dataset):
        self.dataset = dataset

    def get_dataset(self):
        return self.dataset

    def get_bare_model(self, net):
        """Get bare model, especially under wrapping with
        DistributedDataParallel or DataParallel.
        """
        if isinstance(net, (DataParallel, DistributedDataParallel)):
            net = net.module
        return net

    def set_train(self):
        for name, net in self.models.items():
            net.train()

    def set_eval(self):
        for name, net in self.models.items():
            net.eval()
     

    # @master_only
    # def save_model(self, epoch, max_count=-1, max_time=2*60*60):
    #     '''保存模型, 每隔max_time保存一次，在max_time时间内最多保存max_count个
    #     :param epoch: int, 当前模型epoch
    #     :param max_count: int, 在max_time时间内最多保存数量
    #     :param max_time: int, 秒, 最大时间间隔保存一次
    #     '''
    #     for name, net in self.models.items():
    #         optimizer = self.optimizers.get(name, None)
    #         self.save_model_optimizer(epoch, name, net, optimizer, max_count, max_time)

    #     for name, net in self.ema_models.items():
    #         self.save_model_optimizer(epoch, name+'_ema', net, None, max_count, max_time)
    
    @master_only
    def write_network(self, path):
        '''将网络结构写入文件
        :param path: str, 文件路径
        '''
        with open(path, 'w') as file:
            for net in self.models:
                file.write(str(net))

    
    @master_only    
    def get_metric_dict(self):
        '''用于显示相关测量指标
        :return: 返回字典类型或None
        '''
        return None
    
    @master_only
    def get_scalar_dict(self):
        '''用于显示loss等相关指标
        :return: 返回字典类型或None
        '''
        return None
    
    @master_only
    def get_img_dict(self):
        '''用于显示训练期间中间结果图片
        :return: 返回字典类型或None
        '''
        return None

    @master_only
    def get_audio_dict(self):
        '''用于显示训练期间中间结果音频
        :return: 返回字典类型或None
        '''
        return None

    @master_only
    def get_val_dict(self):
        '''用于获取验证集效果
        :return: 返回字典类型或None
        '''
        return None

 



import torch
from torchvision import models
class VGG19_torch(torch.nn.Module):
    def __init__(self, requires_grad=False):
        super(VGG19_torch, self).__init__()
        vgg_pretrained_features = models.vgg19(pretrained=True).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        for x in range(2):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(2, 7):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(7, 12):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(12, 21):
            self.slice4.add_module(str(x), vgg_pretrained_features[x])
        for x in range(21, 30):
            self.slice5.add_module(str(x), vgg_pretrained_features[x])
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        h_relu1 = self.slice1(X)
        h_relu2 = self.slice2(h_relu1)
        h_relu3 = self.slice3(h_relu2)
        h_relu4 = self.slice4(h_relu3)
        h_relu5 = self.slice5(h_relu4)
        out = [h_relu1, h_relu2, h_relu3, h_relu4, h_relu5]
        return out



import json
import requests
import numpy as np
from torchelper.models.model_builder import ModelBuilder
from .callback import Callback
from torchelper.models.base_model import BaseModel
from torchelper.utils.dist_util import master_only, get_bare_model
from torchelper.utils import logger
import os

class WechatCallback(Callback):
    def __init__(self, key):
        super().__init__()
        self.url = "https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key="+key
        self.key = key
     
    
    def send_msg(self, data):
        try:
            post_data = json.dumps(data)
            response = requests.post(self.url,  data=post_data)
        except requests.exceptions.HTTPError as exc:
            print(f"发送失败， HTTP error:{exc.response.status_code} , 原因: {exc.response.reason}")
    
        except requests.exceptions.ConnectionError:
            print("发送失败，HTTP connection error!")
        except requests.exceptions.Timeout:
            print("发送失败，Timeout error!")
            raise
        except requests.exceptions.RequestException:
            print("发送失败, Request Exception!")
            raise
        else:
            result = None
            try:
                result = response.json()
            except json.decoder.JSONDecodeError:
                print(f"服务器响应异常，状态码：{response.status_code}，响应内容：{response.text}")
    
            finally:
                return result 
 

    def get_media_id(self, path, msgtype):
        """上传资源到企业微信的存储上,msgtype有image,voice,video,file"""
        media_url = "https://qyapi.weixin.qq.com/cgi-bin/webhook/upload_media?key={}&type={}".format(self.key, msgtype)
        try:
            with open(path, 'rb') as f:
                files = {msgtype: f}
                r = requests.post(media_url, files=files)
                re = json.loads(r.text)
                id = re.get('media_id', None)
                if id is None:
                    print(re)
                return id
        except:
            return None
    
    def get_media_content_id(self, name, content, msgtype):
        tmp_dir_path = os.path.join(os.path.expanduser('~'), 'tmp')
        if not os.path.exists(tmp_dir_path):
            os.makedirs(tmp_dir_path)
        if msgtype=="voice":
            import wave
            dst = os.path.join(tmp_dir_path, name)
            MIN_VOL = -32768
            MAX_VOL = 32767
            if content.dtype!=np.int16:
                content = (content * np.abs(MIN_VOL)).clip(MIN_VOL, MAX_VOL).astype(np.int16)
            f = wave.open(dst, "wb")
            # 配置声道数、量化位数和取样频率
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(16000)
            f.writeframes(content.tostring())
            f.close()
            id = self.get_media_id(dst, 'file')
            os.remove(dst)
            return id
        
    def send_dict_msg(self, msg:dict, msg_type):
        if msg_type == 'text':
            data = []
            for k, v in msg.items():
                data.append("<font color=\"warning\">" + k + "</font> : " + str(v))
            data = '\n'.join(data)
            self.send_txt(data)
        elif msg_type == 'voice':
            for k, v in msg.items():
               self.send_voice(k, v) 
        elif msg_type == "file":
            data = []
            for k, v in msg.items():
               self.send_file(v)
        else:
            logger.warn("Unrecognized type " + str(msg_type))

    def send_txt(self, msg_str):
        msg = str(msg_str)
        data = {
                "msgtype": "markdown",
                "markdown": {
                        "content": msg
                    }
                }
        self.send_msg(data)

    def send_voice(self, name, voice_content):
        id = self.get_media_content_id(name, voice_content, 'voice')
        if id is None:
            return
        data = {
                "msgtype": "file",
                "file":{
                    "media_id": id,
                }
               }
        self.send_msg(data)

    def send_file(self, file_path, type='file'):
        id = self.get_media_id(file_path, msgtype=type)
        if id is None:
            return
        data = {
                "msgtype": type,
                "file":{
                    "media_id": id,
                }
               }
        self.send_msg(data)

    def on_begin_train(self, model:BaseModel, epoch:int):
        pass

    def on_end_train(self, model:BaseModel):
        pass

    @master_only
    def on_begin_epoch(self, model:BaseModel, epoch:int):
        pass
         
    @master_only
    def send_train_msg(self, builder:ModelBuilder):
        if builder is not None:
            dic = builder.get_scalar_dict()
            if dic is not None:
                self.send_dict_msg(dic, "text")
            dic = builder.get_audio_dict()
            if dic is not None:
                self.send_dict_msg(dic, "voice")

    @master_only
    def on_end_epoch(self, model:BaseModel, epoch:int):
        builder:ModelBuilder = model.get_builder()
        self.send_train_msg(builder)

    def on_begin_step(self, model:BaseModel, epoch:int, step:int):
        pass

    @master_only
    def on_end_step(self, model:BaseModel, epoch:int, step:int):
        pass
        # builder:ModelBuilder = model.get_builder()
        # self.send_train_msg(builder)
        # builder:ModelBuilder = model.get_builder()
        # if builder is not None:
        #     dic = builder.get_scalar_dict()
        #     self.send_dict_msg(dic, "text")


'''读取配置文件
'''
import os
import yaml
import sys
def load_default():
    return {
        'rank': 0,
        'seed': 0,
        'port': 22021,
        'amp' : True,
        'gpu_ids': '0'
    }

def replace_home_dir(cfg):
    home = os.path.expanduser("~")
    if isinstance(cfg, str) and cfg.startswith('~/'):
        return os.path.join(home, cfg[2:])
    elif isinstance(cfg, list) or isinstance(cfg, tuple):
        new_v = []
        for v in cfg:
            v = replace_home_dir(v)
            new_v.append(v)
        if isinstance(cfg, tuple):
            new_v = tuple(new_v)
        return new_v
    elif isinstance(cfg, dict): 
        for key, val in cfg.items():
            val = replace_home_dir(val)
            cfg[key] = val
        return cfg 
    return cfg
    


def load_cfg(path):
    '''加载配置参数
    '''
    file_path = os.path.abspath(__file__)
    file_dir = os.path.dirname(file_path)
    # with open(os.path.join(file_dir, '../default.yml'), 'r', encoding='utf-8') as file:
    #     base_cfg = yaml.safe_load(file)
    base_cfg = load_default()
    with open(path, 'r', encoding='utf-8') as file:
        cfg = yaml.safe_load(file)
    base_cfg.update(cfg)
    base_cfg = replace_home_dir(base_cfg)
    return base_cfg


def parse_args():
    '''解析命令行参数
    '''
    if len(sys.argv) < 2:
        print('Please pass in the yml file, the example is as follows:')
        print('  python train.py your_cfg.yml')
        exit()
    path = sys.argv[1]
    return load_cfg(path)

def init_cfg():
    '''读取并初始化配置文件
    :return: dict
    '''
    cfg = parse_args()
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = cfg['gpu_ids']
    cfg['ori_gpu_ids'] = [int(gpu_id.strip()) for gpu_id in cfg['gpu_ids'].split(",")]
    cfg['gpu_ids'] = [i for i in range(len(cfg['gpu_ids'].split(",")))]
    # check path
    if not os.path.exists( cfg['ckpt_dir']+"/train_out"):
        os.makedirs(cfg['ckpt_dir']+"/train_out")
    return cfg

def merge_cfg(cfg, default_cfg):
    tmp = default_cfg.copy()
    tmp.update(cfg)
    return tmp


import torch
import torch.nn as nn
from torchelper.utils.dist_util import get_rank


class BaseModel(nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.builder = None
        self.__optimizer = None
        self.cfg = cfg
        self.init_lr = cfg['init_lr']
        self.callbacks = []
        self.gpu_id = get_rank()
        if self.gpu_id>=0:
            self.device = torch.device('cuda:'+str(self.gpu_id))
        else:
            self.device = torch.device('cpu')
    def get_builder(self):
        return self.builder
        
    def remap_weights_name(self, weights):
        return weights

    def add_cb(self, cb):
        self.callbacks.append(cb)
    
    def perform_cb(self, event:str, **args):
        for cb in self.callbacks:
            evt_func = getattr(cb, event)
            evt_func(self, **args)

    def get_train_params(self):
        '''返回当前model训练参数，如果返回None，则默认为所有训练参数
        '''
        return None
    
    def get_loss(self):
        '''返回当前model loss，如果返回None，则不做梯度更新
        '''
        return None

    def get_optimizer(self):
        if self.__optimizer is None:
            params = self.get_train_params()
            if params is None:
                params = list(self.parameters())
            if len(params)>0:
                self.__optimizer = torch.optim.Adam(params, lr=self.init_lr, betas=(0.5, 0.999))
        return self.__optimizer




import os
import random
from torchelper.utils.dist_util import get_rank
from torchelper.models.model_builder import ModelBuilder
from tqdm import tqdm
import torch 
import time

import torch.distributed as dist
import torch.multiprocessing as mp
from torchelper.utils.config import merge_cfg
from torchelper.utils.cls_utils import get_cls, new_cls, auto_new_cls
from torchelper.data.base_dataset import get_data_loader
import torch.backends.cudnn as cudnn
import subprocess
from torch.utils.tensorboard import SummaryWriter
from torchelper.utils import logger
 

def check_close_port(port):
    result = subprocess.run(['lsof', '-i:'+str(port)], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return
    print(out)
    lines = lines[1:]
    for line in lines:
        arr = [s for s in line.split(' ') if len(s)>0]
        if len(arr)<2:
            continue
        pid = int(arr[1])
        os.system('kill '+str(pid))
        print("kill", pid)

def check_close_gpu_ids(gpu_ids):
    if not isinstance(gpu_ids, list):
        gpu_ids = [gpu_ids]
    print('kill:', gpu_ids)
    result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return

    start_flag = False
    for line in lines:
        if not start_flag:
            if line.startswith('|=') and not '+' in line:
                start_flag = True
        else:
            line = ' '.join(line.split())
            arr = line.split(' ')
            if len(arr)<8:
                continue
            gpu_id = int(arr[1])
            pid = int(arr[4])
            for gid in gpu_ids:
                if gpu_id == gid and arr[6].endswith('/python'):
                    os.system('kill '+str(pid))
                    print('kill', arr[6])
      

def get_port(def_port):
    result = subprocess.run(['lsof', '-i:'+str(def_port)], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return def_port
    return get_port(def_port+1)


def validate(builder:ModelBuilder, epoch:int, val_dataset):
    '''执行验证集
    :param net: ModelBuilder子类实例
    :param epoch: int, 当前epoch
    :param val_dataset: 验证集
    :return: 返回描述
    :raises ValueError: 描述抛出异常场景
    '''
    if get_rank()==0:
        builder.set_eval()
        # net.before_validate(epoch)
        # time.sleep(10)
        
        val_data:dict = {}
        
        for data in tqdm(val_dataset):
            res = builder.validate(epoch, data)
            if res is None:
                continue
            for key, val in res.items():
                val_data[key] = val + val_data.get(key, 0)
        
        line = 'epoch: '+str(epoch)+', '
        avg_val_dict = {}
        if val_data is not None:
            val_arr = []
            for key, val in val_data.items():
                val_arr.append(key + ":" + str(val_data[key] * 1.0 / len(val_dataset)))
                avg_val_dict[key] = val_data[key] * 1.0 / len(val_dataset)
            line = line+', '.join(val_arr)
        # builder.on_val_end(avg_val_dict)
        builder.perform_cb('on_end_val', epoch=epoch, metric_dict=avg_val_dict)
        print(line)
    torch.distributed.barrier()

#更新tensorboard显示
def update_tensorboard(builder:ModelBuilder, tb_writer, epoch, step, step_per_epoch_per_gpu, gpu_count):
    if tb_writer is None:
        return
    audios = builder.get_audio_dict()
    images = builder.get_img_dict()
    scalars = builder.get_scalar_dict()
    metrics = builder.get_metric_dict()
    audios = audios if audios is not None else {}
    images = images if images is not None else {}
    scalars = scalars if scalars is not None else {}
    metrics = metrics if metrics is not None else {}
 
    step =  (epoch*step_per_epoch_per_gpu+step)*gpu_count
    # print(epoch, step_per_epoch_per_gpu, epoch*step_per_epoch_per_gpu)
    for k, v in audios.items():
        if v is None:
            logger.warn(k+" is None ... ")
            continue    
        tb_writer.add_audio(k, v, step, sample_rate=16000)
    for k, v in images.items():
        if v is None:
            logger.warn(k+" is None ... ")
            continue
        tb_writer.add_image(k, v, step)
    for k, v in scalars.items():
        if v is None:
            logger.warn(k+" is None ... ")
            continue
        tb_writer.add_scalar(k, v, step)
    for k, v in metrics.items():
        if v is None:
            logger.warn(k+" is None ... ")
            continue
        tb_writer.add_scalar(k, v, step)

def update_pbar(builder:ModelBuilder, pbar):
    if pbar is None:
        return
    scalars = builder.get_scalar_dict()
    if scalars is None:
        return
    msg = []
    for k, v in scalars.items():
        msg.append('%s:%.5f'%(k, v))
    pbar.set_description(', '.join(msg))

def train(gpu_id, cfg, is_dist):
   
    train_data_cfg = merge_cfg(cfg['dataset']['train'], cfg)
    val_data_cfg =  merge_cfg(cfg['dataset']['val'], cfg)
    train_dataset = auto_new_cls(train_data_cfg)
    val_dataset = auto_new_cls(val_data_cfg)
    gpu_count = len(cfg['ori_gpu_ids'])
 
    train_dataloader = get_data_loader(cfg['batch_per_gpu'], train_dataset, num_workers=0, dist=True)
    val_dataloader = get_data_loader(cfg['batch_per_gpu'], val_dataset, dist=False)

    builder:ModelBuilder = get_cls(cfg['model_builder'])(cfg, True, cfg['ckpt_dir'])
 
    builder.set_dataset(train_dataset)
    dataset_size = len(train_dataloader)
    
    tb_writer = None
    step_per_epoch_per_gpu = dataset_size 
    if gpu_id==0:
        tb_writer = SummaryWriter(cfg['ckpt_dir'])

    builder.perform_cb('on_begin_train', epoch=cfg['start_epoch'])
    for epoch in range(cfg['start_epoch'], cfg['total_epoch']):
        # validate(builder, epoch, val_dataloader)
        builder.set_train()
        if gpu_id==0:
            pbar = tqdm(train_dataloader)
            enum_data = enumerate(pbar)
        else:
            pbar = None
            enum_data =  enumerate(train_dataloader)
        builder.perform_cb('on_begin_epoch', epoch=epoch)
        for i, data in enum_data:
            builder.on_begin_forward(data, epoch, i)
            builder.perform_cb('on_begin_step', epoch=epoch, step=i)
            builder.forward_wrapper(epoch, i, data)
            builder.on_end_forward(  epoch, i)
            builder.on_begin_backward( epoch, i)
            # 计算loss
            builder.backward_wrapper()
            if is_dist:   # 多卡同步
                torch.distributed.barrier()
            builder.on_end_backward( epoch, i)
            builder.perform_cb('on_end_step', epoch=epoch, step=i)
            if gpu_id==0:
                update_pbar(builder, pbar)
                update_tensorboard(builder, tb_writer, epoch, i, step_per_epoch_per_gpu, gpu_count)
        builder.perform_cb('on_end_epoch', epoch=epoch)
        # builder.save_model(epoch, save_max_count, save_max_time)
        validate(builder, epoch, val_dataloader)
    builder.perform_cb('on_end_train')


def train_worker(gpu_id, nprocs, cfg, is_dist, port):
    '''独立进程运行
    '''
    os.environ['NCCL_BLOCKING_WAIT']="1"
    os.environ['NCCL_ASYNC_ERROR_HANDLING']='1'
    random.seed(0)
    torch.manual_seed(0)
    cudnn.deterministic = True
    # 提升速度，主要对input shape是固定时有效，如果是动态的，耗时反而慢
    torch.backends.cudnn.benchmark = True
    dist.init_process_group(backend='nccl',
                            init_method='tcp://127.0.0.1:'+str(port),
                            world_size=len(cfg['gpu_ids']),
                            rank=gpu_id)

    torch.cuda.set_device(gpu_id)
    # 按batch分割给各个GPU
    # cfg['batch_size'] = int(cfg['batch_size'] / nprocs)
    train(gpu_id, cfg, is_dist)

def train_main(cfg):
    check_close_gpu_ids(cfg['ori_gpu_ids'])
    # check_close_port(cfg['port'])
    gpu_nums = len(cfg['gpu_ids'])
    # if gpu_nums>1:
    port = get_port(cfg['port'])
    print('init port:', port)
    mp.spawn(train_worker, nprocs=gpu_nums, args=(gpu_nums, cfg, True, port))
    # else:
    #     train(cfg['gpu_ids'][0], cfg, False)



import functools
import torch.distributed as dist
from torch.nn.parallel import DataParallel, DistributedDataParallel

def get_rank():
    rank = 0
    if dist.is_available():
        if dist.is_initialized():
            rank = dist.get_rank()
    return rank

def master_only(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        rank = get_rank()
        if rank == 0:
            return func(*args, **kwargs)

    return wrapper

def get_bare_model(net):
    """Get bare model, especially under wrapping with
    DistributedDataParallel or DataParallel.
    """
    if isinstance(net, (DataParallel, DistributedDataParallel)):
        net = net.module
    return net


from .models.model_builder import ModelBuilder
from .callbacks.callback import Callback
from .callbacks.ckpt_callback import CkptCallback
from .callbacks.wechat_send_msg_callback import WechatCallback
from .callbacks.reduce_lr_on_plateau import ReduceLROnPlateau
from .models.base_model import BaseModel
from .utils.config import init_cfg, load_cfg, merge_cfg
from .train import train_main
from .metrics import measure
from .data import *
from .models.lr_scheduler import LinearDownLR
from .utils.dist_util import master_only, get_rank
from .metrics  import *
from .utils.cls_utils import new_cls 
from .utils import logger

name = "torchelper"


# modify from https://github.com/rosinality/stylegan2-pytorch/blob/master/op/fused_act.py # noqa:E501

import os
import torch
from torch import nn
from torch.autograd import Function

 
try:
    from . import fused_act_ext
except ImportError:
    from torch.utils.cpp_extension import load
    module_path = os.path.dirname(__file__)
    fused_act_ext = load(
        'fused',
        sources=[
            os.path.join(module_path, 'src', 'fused_bias_act.cpp'),
            os.path.join(module_path, 'src', 'fused_bias_act_kernel.cu'),
        ],
    )
    # avoid annoying print output
    # print(f'Cannot import deform_conv_ext. Error: {error}. You may need to: \n '
    #       '1. compile with BASICSR_EXT=True. or\n '
    #       '2. set BASICSR_JIT=True during running')


class FusedLeakyReLUFunctionBackward(Function):

    @staticmethod
    def forward(ctx, grad_output, out, negative_slope, scale):
        ctx.save_for_backward(out)
        ctx.negative_slope = negative_slope
        ctx.scale = scale

        empty = grad_output.new_empty(0)

        grad_input = fused_act_ext.fused_bias_act(grad_output, empty, out, 3, 1, negative_slope, scale)

        dim = [0]

        if grad_input.ndim > 2:
            dim += list(range(2, grad_input.ndim))

        grad_bias = grad_input.sum(dim).detach()

        return grad_input, grad_bias

    @staticmethod
    def backward(ctx, gradgrad_input, gradgrad_bias):
        out, = ctx.saved_tensors
        gradgrad_out = fused_act_ext.fused_bias_act(gradgrad_input, gradgrad_bias, out, 3, 1, ctx.negative_slope,
                                                    ctx.scale)

        return gradgrad_out, None, None, None


class FusedLeakyReLUFunction(Function):

    @staticmethod
    def forward(ctx, input, bias, negative_slope, scale):
        empty = input.new_empty(0)
        out = fused_act_ext.fused_bias_act(input, bias, empty, 3, 0, negative_slope, scale)
        ctx.save_for_backward(out)
        ctx.negative_slope = negative_slope
        ctx.scale = scale

        return out

    @staticmethod
    def backward(ctx, grad_output):
        out, = ctx.saved_tensors

        grad_input, grad_bias = FusedLeakyReLUFunctionBackward.apply(grad_output, out, ctx.negative_slope, ctx.scale)

        return grad_input, grad_bias, None, None


class FusedLeakyReLU(nn.Module):

    def __init__(self, channel, negative_slope=0.2, scale=2**0.5):
        super().__init__()

        self.bias = nn.Parameter(torch.zeros(channel))
        self.negative_slope = negative_slope
        self.scale = scale

    def forward(self, input):
        return fused_leaky_relu(input, self.bias, self.negative_slope, self.scale)


def fused_leaky_relu(input, bias, negative_slope=0.2, scale=2**0.5):
    return FusedLeakyReLUFunction.apply(input, bias, negative_slope, scale)




import cv2
import numpy as np
import torch

class Metric():
    def __init__(self):
        pass

    def cal(self, inp, tar):
        pass
    
  

    def img_to_batch_tensor(self, data, device):
        if isinstance(data, np.ndarray):
            if len(data.shape)==4: # 如果是4维，则认为是[b, h, w, c]
                data = np.transpose(data, (0, 3, 1, 2)).astype(np.float32)
                return  torch.from_numpy(data).to(device)
            elif len(data.shape)==3: # 如果是3维，则认为是[h, w, c]
                data = np.expand_dims(data, 0)
                data = np.transpose(data, (0, 3, 1, 2)).astype(np.float32)
                return  torch.from_numpy(data).to(device)
        raise Exception("Unknow type "+str(type(data)))
    
    def align_size(self, uint8_a, uint8_b):
        height, width = uint8_a.shape[:2]
        uint8_b = cv2.resize(uint8_b, (width, height))
        return uint8_a, uint8_b






# from skimage.metrics import peak_signal_noise_ratio as compare_ssim  
from .measure import Metric
import numpy as np
import cv2
import torch.nn as nn
import torch
from math import exp
import torch.nn.functional as F
from torch.autograd import Variable

def gaussian(window_size, sigma):
    gauss = torch.Tensor([exp(-(x - window_size // 2) ** 2 / float(2 * sigma ** 2)) for x in range(window_size)])
    return gauss / gauss.sum()


def create_window(window_size, channel):
    
    window_1d = gaussian(window_size, 1.5).unsqueeze(1)
    window_2d = window_1d.mm(window_1d.t()).float().unsqueeze(0).unsqueeze(0)
    window = Variable(window_2d.expand(channel, 1, window_size, window_size).contiguous())
    return window


def _ssim(img1, img2, window, window_size, channel, size_average=True):
    mu1 = F.conv2d(img1, window, padding=window_size // 2, groups=channel)
    mu2 = F.conv2d(img2, window, padding=window_size // 2, groups=channel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = F.conv2d(img1 * img1, window, padding=window_size // 2, groups=channel) - mu1_sq
    sigma2_sq = F.conv2d(img2 * img2, window, padding=window_size // 2, groups=channel) - mu2_sq
    sigma12 = F.conv2d(img1 * img2, window, padding=window_size // 2, groups=channel) - mu1_mu2

    # C1 = 0.01 ** 2
    # C2 = 0.03 ** 2

    C1 = (0.01 * 255)**2
    C2 = (0.03 * 255)**2

    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))

    if size_average:
        return ssim_map.mean()
    else:
        return ssim_map.mean(1).mean(1).mean(1)





def ssim(img1, img2, window_size=11, size_average=True):
    (_, channel, _, _) = img1.size()
    window = create_window(window_size, channel)

    if img1.is_cuda:
        window = window.cuda(img1.get_device())
    window = window.type_as(img1)

    return _ssim(img1, img2, window, window_size, channel, size_average)





class SSIM(Metric):
    def __init__(self):
        super().__init__()

    def cal(self, inp, tar):
        if len(inp)==3:
            inp, tar = np.array([inp]), np.array([tar])
        
        inp = inp.transpose((0, 3, 1, 2)).astype(np.float64)
        tar = tar.transpose((0, 3, 1, 2)).astype(np.float64)
        inp = torch.from_numpy(inp)
        tar = torch.from_numpy(tar)
        res = ssim(inp, tar)
        res = res.detach().cpu().numpy()
        return res




class SSIMNet(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True):
        super().__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = 1
        self.window = create_window(window_size, self.channel)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.data.type() == img1.data.type():
            window = self.window
        else:
            window = create_window(self.window_size, channel)

            if img1.is_cuda:
                window = window.cuda(img1.get_device())
            window = window.type_as(img1)

            self.window = window
            self.channel = channel

        return _ssim(img1, img2, window, self.window_size, channel, self.size_average)

 
# class SSIM(nn.Module):
#     def __init__(self):
#         super().__init__()

#     def _ssim(self, img1, img2):
#         """Calculate SSIM (structural similarity) for one channel images.
#         It is called by func:`calculate_ssim`.
#         Args:
#             img1 (ndarray): Images with range [0, 255] with order 'HWC'.
#             img2 (ndarray): Images with range [0, 255] with order 'HWC'.
#         Returns:
#             float: ssim result.
#         """

#         C1 = (0.01 * 255)**2
#         C2 = (0.03 * 255)**2

#         img1 = img1.astype(np.float64)
#         img2 = img2.astype(np.float64)
#         kernel = cv2.getGaussianKernel(11, 1.5)
#         window = np.outer(kernel, kernel.transpose())

#         mu1 = cv2.filter2D(img1, -1, window)[5:-5, 5:-5]
#         mu2 = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
#         mu1_sq = mu1**2
#         mu2_sq = mu2**2
#         mu1_mu2 = mu1 * mu2
#         sigma1_sq = cv2.filter2D(img1**2, -1, window)[5:-5, 5:-5] - mu1_sq
#         sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq
#         sigma12 = cv2.filter2D(img1 * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

#         ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))
#         return ssim_map.mean()

    
#     def calculate_ssim(self, img1, img2):
#         """Calculate SSIM (structural similarity).
#         Ref:
#         Image quality assessment: From error visibility to structural similarity
#         The results are the same as that of the official released MATLAB code in
#         https://ece.uwaterloo.ca/~z70wang/research/ssim/.
#         For three-channel images, SSIM is calculated for each channel and then
#         averaged.
#         Args:
#             img1 (ndarray): Images with range [0, 255].
#             img2 (ndarray): Images with range [0, 255].
#         Returns:
#             float: ssim result.
#         """
#         img1 = img1.astype(np.float64)
#         img2 = img2.astype(np.float64)
#         ssims = []
#         for i in range(img1.shape[2]):
#             ssims.append(self._ssim(img1[..., i], img2[..., i]))
#         return np.array(ssims).mean()

#     # def ssim(self, img_a, img_b):
#         # '''测量ssim
#         # :param img_a: np.array, dtype=np.uint8, 输入图
#         # :param img_b: np.array, dtype=np.uint8, GT
#         # :return: 返回ssim值
#         # '''
#         # # multichannel: If True, treat the last dimension of the array as channels.
#         # # Similarity calculations are done independently for each channel then averaged.
#         # print(img_a.shape)
#         # print(img_b.shape)
#         # score, __ = compare_ssim(img_a, img_b, full=True, multichannel=True)
#         # return score

#     def cal(self, inp, tar):
#         if len(inp)==3:
#             batch_size = 1
#             inp, tar = [inp], [tar]
#         else:
#             batch_size = inp.shape[0]
#         sum_v = 0
#         for i in range(batch_size):
#             sum_v += self.calculate_ssim(inp[i], tar[i])
#         res = sum_v * 1.0 / batch_size
#         print(res)
#         return res



#!/usr/bin/env python3
"""Calculates the Frechet Inception Distance (FID) to evalulate GANs

The FID metric calculates the distance between two distributions of images.
Typically, we have summary statistics (mean & covariance matrix) of one
of these distributions, while the 2nd distribution is given by a GAN.

When run as a stand-alone program, it compares the distribution of
images that are stored as PNG/JPEG at a specified location with a
distribution given by summary statistics (in pickle format).

The FID is calculated by assuming that X_1 and X_2 are the activations of
the pool_3 layer of the inception net for generated samples and real world
samples respectively.

See --help to see further details.

Code apapted from https://github.com/bioinf-jku/TTUR to use PyTorch instead
of Tensorflow

Copyright 2018 Institute of Bioinformatics, JKU Linz

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import pathlib
from tqdm import tqdm

import cv2
import numpy as np
import torch
from scipy import linalg
from torch.nn.functional import adaptive_avg_pool2d

from torchelper.metrics.measure import Metric
from torchelper.archs.inception import InceptionV3

# try:
#     from tqdm import tqdm
# except ImportError:
#     # If not tqdm is not available, provide a mock version of it
#     def tqdm(x):
#         return x

# parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
# parser.add_argument('path', type=str, nargs=2,
#                     help=('Path to the generated images or '
#                           'to .npz statistic files'))
# parser.add_argument('--batch-size', type=int, default=50,
#                     help='Batch size to use')
# parser.add_argument('--dims', type=int, default=2048,
#                     choices=list(InceptionV3.BLOCK_INDEX_BY_DIM),
#                     help=('Dimensionality of Inception features to use. '
#                           'By default, uses pool3 features'))
# parser.add_argument('-c', '--gpu', default='', type=str,
#                     help='GPU to use (leave blank for CPU only)')


# def get_activations(files, model, batch_size=50, dims=2048,
#                     cuda=False, verbose=False, use_tqdm=False):
#     """Calculates the activations of the pool_3 layer for all images.

#     Params:
#     -- files       : List of image files paths
#     -- model       : Instance of inception model
#     -- batch_size  : Batch size of images for the model to process at once.
#                      Make sure that the number of samples is a multiple of
#                      the batch size, otherwise some samples are ignored. This
#                      behavior is retained to match the original FID score
#                      implementation.
#     -- dims        : Dimensionality of features returned by Inception
#     -- cuda        : If set to True, use GPU
#     -- verbose     : If set to True and parameter out_step is given, the number
#                      of calculated batches is reported.
#     Returns:
#     -- A numpy array of dimension (num images, dims) that contains the
#        activations of the given tensor when feeding inception with the
#        query tensor.
#     """
#     model.eval()

#     if len(files) % batch_size != 0:
#         print(('Warning: number of images is not a multiple of the '
#                'batch size. Some samples are going to be ignored.'))
#     if batch_size > len(files):
#         print(('Warning: batch size is bigger than the datasets size. '
#                'Setting batch size to datasets size'))
#         batch_size = len(files)
#     # print(len(files), batch_size)

#     n_batches = len(files) // batch_size
#     n_used_imgs = n_batches * batch_size

#     pred_arr = np.empty((n_used_imgs, dims))

#     if use_tqdm:
#         from tqdm import tqdm
#     else:
#         def tqdm(x):
#             return x
#     for i in tqdm(range(n_batches)):
#         if verbose:
#             print('\rPropagating batch %d/%d' % (i + 1, n_batches),
#                   end='', flush=True)
#         start = i * batch_size
#         end = start + batch_size
#         images = np.array([imread(str(f)).astype(np.float32)
#                            for f in files[start:end]])
#         if len(images.shape) != 4:
#             images = imread(str(files[start]))
#             images = cv2.cvtColor(images, cv2.COLOR_BGR2GRAY)
#             images = np.array([images.astype(np.float32)])
#         # Reshape to (n_images, 3, height, width)
#         images = images.transpose((0, 3, 1, 2))
#         images /= 255

#         batch = torch.from_numpy(images).type(torch.FloatTensor)
#         if cuda:
#             batch = batch.cuda()
#         with torch.no_grad():
#             pred = model(batch)[0]

#         # If model output is not scalar, apply global spatial average pooling.
#         # This happens if you choose a dimensionality not equal 2048.

#         if pred.shape[2] != 1 or pred.shape[3] != 1:
#             pred = adaptive_avg_pool2d(pred, output_size=(1, 1))

#         pred_arr[start:end] = pred.cpu().data.numpy().reshape(end - start, -1)

#     if verbose:
#         print(' done')

#     return pred_arr


# def get_activations_from_ims(ims, model, batch_size=50, dims=2048,
#                              device=None, verbose=False, tqdm_position=None):
#     """Calculates the activations of the pool_3 layer for all images.

#     Params:
#     -- files       : List of image files paths
#     -- model       : Instance of inception model
#     -- batch_size  : Batch size of images for the model to process at once.
#                      Make sure that the number of samples is a multiple of
#                      the batch size, otherwise some samples are ignored. This
#                      behavior is retained to match the original FID score
#                      implementation.
#     -- dims        : Dimensionality of features returned by Inception
#     -- cuda        : If set to True, use GPU
#     -- verbose     : If set to True and parameter out_step is given, the number
#                      of calculated batches is reported.
#     Returns:
#     -- A numpy array of dimension (num images, dims) that contains the
#        activations of the given tensor when feeding inception with the
#        query tensor.
#     """
#     model.eval()

#     # if len(ims) % batch_size != 0:
#     #     print(('Warning: number of images is not a multiple of the '
#     #            'batch size. Some samples are going to be ignored.'))
#     # if batch_size > len(ims):
#     #     print(('Warning: batch size is bigger than the datasets size. '
#     #            'Setting batch size to datasets size'))
#     #     batch_size = len(ims)
#     # print(len(files), batch_size)

#     n_batches = (len(ims) + batch_size - 1) // batch_size
#     n_used_imgs = len(ims)

#     pred_arr = np.empty((n_used_imgs, dims))

#     if tqdm_position is None or tqdm_position >= 0:
#         import tqdm
#         dataloader_tqdm = tqdm.trange(n_batches, desc='FID        ', position=tqdm_position, leave=False)
#     else:
#         dataloader_tqdm = range(n_batches)
#     for i in dataloader_tqdm:
#         start = i * batch_size
#         end = start + batch_size
#         if end > len(ims):
#             end = len(ims)
#         images = ims[start:end]
#         if images.shape[1] != 3:
#             images = images.transpose((0, 3, 1, 2))
#         images /= 255

#         batch = torch.from_numpy(images).type(torch.FloatTensor).to(device)
#         # if cuda:
#         #     batch = batch.cuda()
#         with torch.no_grad():
#             pred = model(batch)[0]

#         # If model output is not scalar, apply global spatial average pooling.
#         # This happens if you choose a dimensionality not equal 2048.
#         # print(pred.shape)
#         if pred.shape[2] != 1 or pred.shape[3] != 1:
#             pred = adaptive_avg_pool2d(pred, output_size=(1, 1))

#         pred_arr[start:end] = pred.cpu().data.numpy().reshape(end - start, -1)

#     if verbose:
#         print(' done')

#     return pred_arr


# def calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
#     """Numpy implementation of the Frechet Distance.
#     The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
#     and X_2 ~ N(mu_2, C_2) is
#             d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).

#     Stable version by Dougal J. Sutherland.

#     Params:
#     -- mu1   : Numpy array containing the activations of a layer of the
#                inception net (like returned by the function 'get_predictions')
#                for generated samples.
#     -- mu2   : The sample mean over activations, precalculated on an
#                representative datasets set.
#     -- sigma1: The covariance matrix over activations for generated samples.
#     -- sigma2: The covariance matrix over activations, precalculated on an
#                representative datasets set.

#     Returns:
#     --   : The Frechet Distance.
#     """

#     mu1 = np.atleast_1d(mu1)
#     mu2 = np.atleast_1d(mu2)

#     sigma1 = np.atleast_2d(sigma1)
#     sigma2 = np.atleast_2d(sigma2)

#     assert mu1.shape == mu2.shape, \
#         'Training and test mean vectors have different lengths'
#     assert sigma1.shape == sigma2.shape, \
#         'Training and test covariances have different dimensions'

#     diff = mu1 - mu2

#     # print(t.max())
#     # print(t.min())
#     # print(abs(t).mean())
#     # print(t.mean())

#     # Product might be almost singular
#     t = sigma1.dot(sigma2)
#     for i in range(30):
#         # print(i)
#         flag = True
#         covmean, _ = linalg.sqrtm(t, disp=False)
#         if not np.isfinite(covmean).all():
#             msg = ('fid calculation produces singular product; '
#                    'adding %s to diagonal of cov estimates') % eps
#             print(msg)
#             offset = np.eye(sigma1.shape[0]) * eps
#             covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

#         # Numerical error might give slight imaginary component
#         if np.iscomplexobj(covmean):
#             if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
#                 # raise ValueError('Imaginary component {}'.format(m))
#                 flag = False
#             covmean = covmean.real
#         if flag:
#             break
#     if not flag:
#         print('Warning: the fid may be incorrect!')
#     tr_covmean = np.trace(covmean)

#     return (diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean)


# def calculate_activation_statistics(files, model, batch_size=50,
#                                     dims=2048, cuda=False, verbose=False, use_tqdm=False):
#     """Calculation of the statistics used by the FID.
#     Params:
#     -- files       : List of image files paths
#     -- model       : Instance of inception model
#     -- batch_size  : The images numpy array is split into batches with
#                      batch size batch_size. A reasonable batch size
#                      depends on the hardware.
#     -- dims        : Dimensionality of features returned by Inception
#     -- cuda        : If set to True, use GPU
#     -- verbose     : If set to True and parameter out_step is given, the
#                      number of calculated batches is reported.
#     Returns:
#     -- mu    : The mean over samples of the activations of the pool_3 layer of
#                the inception model.
#     -- sigma : The covariance matrix of the activations of the pool_3 layer of
#                the inception model.
#     """
#     act = get_activations(files, model, batch_size, dims, cuda, verbose, use_tqdm=use_tqdm)
#     mu = np.mean(act, axis=0)
#     sigma = np.cov(act, rowvar=False)
#     return mu, sigma


# def _compute_statistics_of_path(path, model, batch_size, dims, cuda, use_tqdm=False):
#     if path.endswith('.npz'):
#         f = np.load(path)
#         m, s = f['mu'][:], f['sigma'][:]
#         f.close()
#     else:
#         path = pathlib.Path(path)
#         files = list(path.glob('*.jpg')) + list(path.glob('*.png'))
#         m, s = calculate_activation_statistics(files, model, batch_size,
#                                                dims, cuda, use_tqdm=use_tqdm)
#     return m, s


# def _compute_statistics_of_ims(ims, model, batch_size, dims, device, tqdm_position=None):
#     act = get_activations_from_ims(ims, model, batch_size, dims, device, verbose=False, tqdm_position=tqdm_position)
#     mu = np.mean(act, axis=0)
#     sigma = np.cov(act, rowvar=False)
#     return mu, sigma


# def calculate_fid_given_ims(ims_fake, ims_real, batch_size, cuda, dims, model=None, tqdm_position=None):
#     if model is None:
#         block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[dims]

#         model = InceptionV3([block_idx])
#         if cuda:
#             model.cuda()

#     m1, s1 = _compute_statistics_of_ims(ims_fake, model, batch_size,
#                                         dims, cuda, tqdm_position=tqdm_position)
#     m2, s2 = _compute_statistics_of_ims(ims_real, model, batch_size,
#                                         dims, cuda, tqdm_position=tqdm_position)

#     fid_value = calculate_frechet_distance(m1, s1, m2, s2)

#     return fid_value


# def calculate_fid_given_paths(paths, batch_size, cuda, dims, model=None, use_tqdm=False):
#     """Calculates the FID of two paths"""
#     for p in paths:
#         if not os.path.exists(p):
#             raise RuntimeError('Invalid path: %s' % p)
#     if model is None:
#         block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[dims]

#         model = InceptionV3([block_idx])
#         if cuda:
#             model.cuda()
#     else:
#         pass

#     m1, s1 = _compute_statistics_of_path(paths[0], model, batch_size,
#                                          dims, cuda, use_tqdm=use_tqdm)
#     m2, s2 = _compute_statistics_of_path(paths[1], model, batch_size,
#                                          dims, cuda, use_tqdm=use_tqdm)
#     fid_value = calculate_frechet_distance(m1, s1, m2, s2)

#     return fid_value

class FID(Metric):
    def __init__(self, tar_dir_or_npz, gpu_id=None, batch_size=20, rect=None, dst_size=None):
        super().__init__()
        self.device = torch.device('cuda:'+str(gpu_id) if gpu_id is not None else 'cpu')
        self.rect = rect
        self.dst_size = dst_size
        self.batch_size = batch_size
        self.dims = 2048
        block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[self.dims]
        model = InceptionV3([block_idx])
        model =  model.to(self.device)
        self.model = model.eval()
        if tar_dir_or_npz.endswith('.npz'):
            npzfile = np.load(open(tar_dir_or_npz, 'rb'))
            self.tar_m = npzfile['m']
            self.tar_s = npzfile['s']
        else:
            self.tar_m, self.tar_s = self.calc_dir(tar_dir_or_npz)
 
    def save_tar(self, dst):
        if not dst.endswith('.npz'):
            dst = dst+'.npz'
        with open(dst, 'wb+') as file:
            np.savez(file, m=self.tar_m, s=self.tar_s)

    def run_model(self, batch):
        
        with torch.no_grad():
            pred = self.model(batch)[0]

        # If model output is not scalar, apply global spatial average pooling.
        # This happens if you choose a dimensionality not equal 2048.
        if pred.shape[2] != 1 or pred.shape[3] != 1:
            pred = adaptive_avg_pool2d(pred, output_size=(1, 1))
        return  pred.cpu().data.numpy()

    def calc_dir(self, dir_path):
        # act = self.get_activations(files)

        names = os.listdir(dir_path)
        batch_size = self.batch_size
        if batch_size > len(names):
            batch_size = len(names)

        n_batches = len(names) // batch_size
        n_used_imgs = n_batches * batch_size

        pred_arr = np.empty((n_used_imgs, self.dims))
        for i in tqdm(range(n_batches)):
            start = i * batch_size
            end = start + batch_size
            images = []
            for name in names[start:end]:
                img = cv2.imread(os.path.join(dir_path, name))
                if self.rect is not None:
                    x1, y1, x2, y2 = self.rect
                    img = img[y1:y2, x1:x2, :]
                if self.dst_size is not None:
                    img = cv2.resize(img, self.dst_size)
                img = img.astype(np.float32)
                images.append(img)
            images = np.array(images)
            # Reshape to (n_images, 3, height, width)
            images = images.transpose((0, 3, 1, 2))
            images /= 255

            batch = torch.from_numpy(images).type(torch.FloatTensor)
            batch = batch.to(self.device)
            
            pred_arr[start:end] = self.run_model(batch).reshape(end - start, -1)

        mu = np.mean(pred_arr, axis=0)
        sigma = np.cov(pred_arr, rowvar=False)
        return mu, sigma
    
    def cal_frechet_distance(self, mu1, sigma1, mu2, sigma2, eps=1e-6):
        """Numpy implementation of the Frechet Distance.
        The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
        and X_2 ~ N(mu_2, C_2) is
                d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).

        Stable version by Dougal J. Sutherland.

        Params:
        -- mu1   : Numpy array containing the activations of a layer of the
                inception net (like returned by the function 'get_predictions')
                for generated samples.
        -- mu2   : The sample mean over activations, precalculated on an
                representative datasets set.
        -- sigma1: The covariance matrix over activations for generated samples.
        -- sigma2: The covariance matrix over activations, precalculated on an
                representative datasets set.

        Returns:
        --   : The Frechet Distance.
        """

        mu1 = np.atleast_1d(mu1)
        mu2 = np.atleast_1d(mu2)

        sigma1 = np.atleast_2d(sigma1)
        sigma2 = np.atleast_2d(sigma2)

        assert mu1.shape == mu2.shape, \
            'Training and test mean vectors have different lengths'
        assert sigma1.shape == sigma2.shape, \
            'Training and test covariances have different dimensions'

        diff = mu1 - mu2

        # print(t.max())
        # print(t.min())
        # print(abs(t).mean())
        # print(t.mean())

        # Product might be almost singular
        t = sigma1.dot(sigma2)
        for i in range(30):
            # print(i)
            flag = True
            covmean, _ = linalg.sqrtm(t, disp=False)
            if not np.isfinite(covmean).all():
                msg = ('fid calculation produces singular product; '
                    'adding %s to diagonal of cov estimates') % eps
                print(msg)
                offset = np.eye(sigma1.shape[0]) * eps
                covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

            # Numerical error might give slight imaginary component
            if np.iscomplexobj(covmean):
                if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
                    # raise ValueError('Imaginary component {}'.format(m))
                    flag = False
                covmean = covmean.real
            if flag:
                break
        if not flag:
            print('Warning: the fid may be incorrect!')
        tr_covmean = np.trace(covmean)

        return (diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean)

    def cal(self, dir_path):
        inp_m, inp_s = self.calc_dir(dir_path)
        fid = self.cal_frechet_distance(inp_m, inp_s, self.tar_m, self.tar_s)
        return fid

def main():
    
    fid = FID('/home/super/Code/FaceEnhance/data/ffhq_m_s.npz', gpu_id=0, batch_size=50, rect=[320*2, 0, 320*3, 320])
    # fid.save_tar('/home/super/Code/FaceEnhance/data/ffhq_m_s.npz')
    fid = fid.cal('/home/super/Model/pix2pix/ffhq_0914/imgs/7')
    print(fid )